# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from ._common import EventData, EventDataBatch
from ._version import VERSION

__version__ = VERSION

from ._constants import TransportType
from ._producer_client import EventHubProducerClient
from ._consumer_client import EventHubConsumerClient
from ._client_base import EventHubSharedKeyCredential
from ._eventprocessor.checkpoint_store import CheckpointStore
from ._eventprocessor.common import CloseReason, LoadBalancingStrategy
from ._eventprocessor.partition_context import PartitionContext
from ._connection_string_parser import (
    parse_connection_string,
    EventHubConnectionStringProperties,
)

__all__ = [
    "EventData",
    "EventDataBatch",
    "EventHubProducerClient",
    "EventHubConsumerClient",
    "TransportType",
    "EventHubSharedKeyCredential",
    "CheckpointStore",
    "CloseReason",
    "LoadBalancingStrategy",
    "PartitionContext",
    "parse_connection_string",
    "EventHubConnectionStringProperties",
]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------

from __future__ import annotations
from typing import Optional, Any, cast, Mapping, Dict, Union, List

from ._amqp_utils import normalized_data_body, normalized_sequence_body
from ._constants import AmqpMessageBodyType
from .._mixin import DictMixin


class AmqpAnnotatedMessage:
    # pylint: disable=too-many-instance-attributes
    """
    The AMQP Annotated Message for advanced sending and receiving scenarios which allows you to
    access to low-level AMQP message sections. There should be one and only one of either data_body, sequence_body
    or value_body being set as the body of the AmqpAnnotatedMessage; if more than one body is set, `ValueError` will
    be raised.
    Please refer to the AMQP spec:
    http://docs.oasis-open.org/amqp/core/v1.0/os/amqp-core-messaging-v1.0-os.html#section-message-format
    for more information on the message format.

    :keyword data_body: The body consists of one or more data sections and each section contains opaque binary data.
    :paramtype data_body: Union[str, bytes, List[Union[str, bytes]]]
    :keyword sequence_body: The body consists of one or more sequence sections and
     each section contains an arbitrary number of structured data elements.
    :paramtype sequence_body: List[Any]
    :keyword value_body: The body consists of one amqp-value section and the section contains a single AMQP value.
    :paramtype value_body: Any
    :keyword header: The amqp message header.
    :paramtype header: Optional[~azure.eventhub.amqp.AmqpMessageHeader]
    :keyword footer: The amqp message footer.
    :paramtype footer: Optional[Dict]
    :keyword properties: Properties to add to the amqp message.
    :paramtype properties: Optional[~azure.eventhub.amqp.AmqpMessageProperties]
    :keyword application_properties: Service specific application properties.
    :paramtype application_properties: Optional[Dict]
    :keyword annotations: Service specific message annotations.
    :paramtype annotations: Optional[Dict]
    :keyword delivery_annotations: Service specific delivery annotations.
    :paramtype delivery_annotations: Optional[Dict]
    """

    def __init__(self, **kwargs: Any) -> None:
        self._encoding = kwargs.pop("encoding", "UTF-8")
        self._data_body: Optional[Union[str, bytes, List[Union[str, bytes]]]] = None
        self._sequence_body: Optional[List[Any]] = None
        self._value_body: Any = None

        # internal usage only for Event Hub received message
        message = kwargs.pop("message", None)
        if message:
            self._from_amqp_message(message)
            return

        # manually constructed AMQPAnnotatedMessage
        input_count_validation = len([key for key in ("data_body", "sequence_body", "value_body") if key in kwargs])
        if input_count_validation != 1:
            raise ValueError(
                "There should be one and only one of either data_body, sequence_body "
                "or value_body being set as the body of the AmqpAnnotatedMessage."
            )

        self._body_type: AmqpMessageBodyType = None  # type: ignore
        if "data_body" in kwargs:
            self._data_body = normalized_data_body(kwargs.get("data_body"))
            self._body_type = AmqpMessageBodyType.DATA
        elif "sequence_body" in kwargs:
            self._sequence_body = normalized_sequence_body(kwargs.get("sequence_body"))
            self._body_type = AmqpMessageBodyType.SEQUENCE
        elif "value_body" in kwargs:
            self._value_body = kwargs.get("value_body")
            self._body_type = AmqpMessageBodyType.VALUE

        header_dict = cast(Mapping, kwargs.get("header"))
        self._header = AmqpMessageHeader(**header_dict) if "header" in kwargs else None
        self._footer = kwargs.get("footer")
        properties_dict = cast(Mapping, kwargs.get("properties"))
        self._properties = AmqpMessageProperties(**properties_dict) if "properties" in kwargs else None
        self._application_properties = kwargs.get("application_properties")
        self._annotations = kwargs.get("annotations")
        self._delivery_annotations = kwargs.get("delivery_annotations")

    def __str__(self) -> str:
        if self._body_type == AmqpMessageBodyType.DATA:
            return "".join(d.decode(self._encoding) for d in self._data_body)  # type: ignore
        if self._body_type == AmqpMessageBodyType.SEQUENCE:
            return str(self._sequence_body)
        if self._body_type == AmqpMessageBodyType.VALUE:
            return str(self._value_body)
        return ""

    def __repr__(self) -> str:
        # pylint: disable=bare-except
        message_repr = "body={}".format(str(self))
        message_repr += ", body_type={}".format(self.body_type.value)
        try:
            message_repr += ", header={}".format(self.header)
        except:
            message_repr += ", header=<read-error>"
        try:
            message_repr += ", footer={}".format(self.footer)
        except:
            message_repr += ", footer=<read-error>"
        try:
            message_repr += ", properties={}".format(self.properties)
        except:
            message_repr += ", properties=<read-error>"
        try:
            message_repr += ", application_properties={}".format(self.application_properties)
        except:
            message_repr += ", application_properties=<read-error>"
        try:
            message_repr += ", delivery_annotations={}".format(self.delivery_annotations)
        except:
            message_repr += ", delivery_annotations=<read-error>"
        try:
            message_repr += ", annotations={}".format(self.annotations)
        except:
            message_repr += ", annotations=<read-error>"
        return "AmqpAnnotatedMessage({})".format(message_repr)[:1024]

    def _from_amqp_message(self, message):
        self._properties = (
            AmqpMessageProperties(
                message_id=message.properties.message_id,
                user_id=message.properties.user_id,
                to=message.properties.to,
                subject=message.properties.subject,
                reply_to=message.properties.reply_to,
                correlation_id=message.properties.correlation_id,
                content_type=message.properties.content_type,
                content_encoding=message.properties.content_encoding,
                absolute_expiry_time=message.properties.absolute_expiry_time,
                creation_time=message.properties.creation_time,
                group_id=message.properties.group_id,
                group_sequence=message.properties.group_sequence,
                reply_to_group_id=message.properties.reply_to_group_id,
            )
            if message.properties
            else None
        )
        self._header = (
            AmqpMessageHeader(
                delivery_count=message.header.delivery_count,
                time_to_live=message.header.ttl,
                first_acquirer=message.header.first_acquirer,
                durable=message.header.durable,
                priority=message.header.priority,
            )
            if message.header
            else None
        )
        self._footer = message.footer if message.footer else {}
        self._annotations = message.message_annotations if message.message_annotations else {}
        self._delivery_annotations = message.delivery_annotations if message.delivery_annotations else {}
        self._application_properties = message.application_properties if message.application_properties else {}
        if message.data:
            self._data_body = cast(List, list(message.data))
            self._body_type = AmqpMessageBodyType.DATA
        elif message.sequence:
            self._sequence_body = cast(List, list(message.sequence))
            self._body_type = AmqpMessageBodyType.SEQUENCE
        else:
            self._value_body = message.value
            self._body_type = AmqpMessageBodyType.VALUE

    @property
    def body(self) -> Any:
        """
        The body of the Message. The format may vary depending on the body type:

        For AmqpMessageBodyType.DATA, the body could be bytes or Iterable[bytes].
        For AmqpMessageBodyType.SEQUENCE, the body could be List or Iterable[List].
        For AmqpMessageBodyType.VALUE, the body could be any type.

        :rtype: Any
        """
        if self._body_type == AmqpMessageBodyType.DATA:  # pylint:disable=no-else-return
            return (i for i in cast(List, self._data_body))  # type: ignore
        elif self._body_type == AmqpMessageBodyType.SEQUENCE:
            return (i for i in cast(List, self._sequence_body))
        elif self._body_type == AmqpMessageBodyType.VALUE:
            return self._value_body
        return None

    @property
    def body_type(self) -> AmqpMessageBodyType:
        """
        The body type of the underlying AMQP message.

        :rtype: ~azure.eventhub.amqp.AmqpMessageBodyType
        """
        return self._body_type

    @property
    def properties(self) -> Optional[AmqpMessageProperties]:
        """
        Properties to add to the message.

        :rtype: Optional[~azure.eventhub.amqp.AmqpMessageProperties]
        """
        return self._properties

    @properties.setter
    def properties(self, value: AmqpMessageProperties) -> None:
        self._properties = value

    @property
    def application_properties(self) -> Optional[Dict[Union[str, bytes], Any]]:
        """
        Service specific application properties.

        :rtype: Optional[Dict]
        """
        return self._application_properties

    @application_properties.setter
    def application_properties(self, value: Optional[Dict[Union[str, bytes], Any]]) -> None:
        self._application_properties = value

    @property
    def annotations(self) -> Optional[Dict[Union[str, bytes], Any]]:
        """
        Service specific message annotations.

        :rtype: Optional[Dict]
        """
        return self._annotations

    @annotations.setter
    def annotations(self, value: Optional[Dict[Union[str, bytes], Any]]) -> None:
        self._annotations = value

    @property
    def delivery_annotations(self) -> Optional[Dict[Union[str, bytes], Any]]:
        """
        Delivery-specific non-standard properties at the head of the message.
        Delivery annotations convey information from the sending peer to the receiving peer.

        :rtype: Dict
        """
        return self._delivery_annotations

    @delivery_annotations.setter
    def delivery_annotations(self, value: Optional[Dict[Union[str, bytes], Any]]) -> None:
        self._delivery_annotations = value

    @property
    def header(self) -> Optional[AmqpMessageHeader]:
        """
        The message header.

        :rtype: Optional[~azure.eventhub.amqp.AmqpMessageHeader]
        """
        return self._header

    @header.setter
    def header(self, value: AmqpMessageHeader) -> None:
        self._header = value

    @property
    def footer(self) -> Optional[Dict[Any, Any]]:
        """
        The message footer.

        :rtype: Optional[Dict]
        """
        return self._footer

    @footer.setter
    def footer(self, value: Optional[Dict[Any, Any]]) -> None:
        self._footer = value


class AmqpMessageHeader(DictMixin):
    """The Message header.
    The Message header. This is only used on received message, and not
    set on messages being sent. The properties set on any given message
    will depend on the Service and not all messages will have all properties.
    Please refer to the AMQP spec:
    http://docs.oasis-open.org/amqp/core/v1.0/os/amqp-core-messaging-v1.0-os.html#type-header
    for more information on the message header.

    :keyword delivery_count: The number of unsuccessful previous attempts to deliver
     this message. If this value is non-zero it can be taken as an indication that the
     delivery might be a duplicate. On first delivery, the value is zero. It is
     incremented upon an outcome being settled at the sender, according to rules
     defined for each outcome.
    :paramtype delivery_count: Optional[int]
    :keyword time_to_live: Duration in milliseconds for which the message is to be considered "live".
     If this is set then a message expiration time will be computed based on the time of arrival
     at an intermediary. Messages that live longer than their expiration time will be discarded
     (or dead lettered). When a message is transmitted by an intermediary that was received
     with a ttl, the transmitted message's header SHOULD contain a ttl that is computed as the
     difference between the current time and the formerly computed message expiration time,
     i.e., the reduced ttl, so that messages will eventually die if they end up in a delivery loop.
    :paramtype time_to_live: Optional[int]
    :keyword durable: Durable messages MUST NOT be lost even if an intermediary is unexpectedly terminated
     and restarted. A target which is not capable of fulfilling this guarantee MUST NOT accept messages
     where the durable header is set to `True`: if the source allows the rejected outcome then the
     message SHOULD be rejected with the precondition-failed error, otherwise the link MUST be detached
     by the receiver with the same error.
    :paramtype durable: Optional[bool]
    :keyword first_acquirer: If this value is `True`, then this message has not been acquired
     by any other link. If this value is `False`, then this message MAY have previously
     been acquired by another link or links.
    :paramtype first_acquirer: Optional[bool]
    :keyword priority: This field contains the relative message priority. Higher numbers indicate higher
     priority messages. Messages with higher priorities MAY be delivered before those with lower priorities.
    :paramtype priority: Optional[int]
    :ivar delivery_count: The number of unsuccessful previous attempts to deliver
     this message. If this value is non-zero it can be taken as an indication that the
     delivery might be a duplicate. On first delivery, the value is zero. It is
     incremented upon an outcome being settled at the sender, according to rules
     defined for each outcome.
    :vartype delivery_count: Optional[int]
    :ivar time_to_live: Duration in milliseconds for which the message is to be considered "live".
     If this is set then a message expiration time will be computed based on the time of arrival
     at an intermediary. Messages that live longer than their expiration time will be discarded
     (or dead lettered). When a message is transmitted by an intermediary that was received
     with a ttl, the transmitted message's header SHOULD contain a ttl that is computed as the
     difference between the current time and the formerly computed message expiration time,
     i.e., the reduced ttl, so that messages will eventually die if they end up in a delivery loop.
    :vartype time_to_live: Optional[int]
    :ivar durable: Durable messages MUST NOT be lost even if an intermediary is unexpectedly terminated
     and restarted. A target which is not capable of fulfilling this guarantee MUST NOT accept messages
     where the durable header is set to `True`: if the source allows the rejected outcome then the
     message SHOULD be rejected with the precondition-failed error, otherwise the link MUST be detached
     by the receiver with the same error.
    :vartype durable: Optional[bool]
    :ivar first_acquirer: If this value is `True`, then this message has not been acquired
     by any other link. If this value is `False`, then this message MAY have previously
     been acquired by another link or links.
    :vartype first_acquirer: Optional[bool]
    :ivar priority: This field contains the relative message priority. Higher numbers indicate higher
     priority messages. Messages with higher priorities MAY be delivered before those with lower priorities.
    :vartype priority: Optional[int]
    """

    def __init__(self, **kwargs):
        self.delivery_count = kwargs.get("delivery_count")
        self.time_to_live = kwargs.get("time_to_live")
        self.first_acquirer = kwargs.get("first_acquirer")
        self.durable = kwargs.get("durable")
        self.priority = kwargs.get("priority")


class AmqpMessageProperties(DictMixin):
    # pylint: disable=too-many-instance-attributes
    """Message properties.
    The properties that are actually used will depend on the service implementation.
    Not all received messages will have all properties, and not all properties
    will be utilized on a sent message.
    Please refer to the AMQP spec:
    http://docs.oasis-open.org/amqp/core/v1.0/os/amqp-core-messaging-v1.0-os.html#type-properties
    for more information on the message properties.

    :keyword message_id: Message-id, if set, uniquely identifies a message within the message system.
     The message producer is usually responsible for setting the message-id in such a way that it
     is assured to be globally unique. A broker MAY discard a message as a duplicate if the value
     of the message-id matches that of a previously received message sent to the same node.
    :paramtype message_id: Optional[Union[str, bytes, uuid.UUID]]
    :keyword user_id: The identity of the user responsible for producing the message. The client sets
     this value, and it MAY be authenticated by intermediaries.
    :paramtype user_id: Optional[Union[str, bytes]]
    :keyword to: The to field identifies the node that is the intended destination of the message.
     On any given transfer this might not be the node at the receiving end of the link.
    :paramtype to: Optional[Union[str, bytes]]
    :keyword subject: A common field for summary information about the message content and purpose.
    :paramtype subject: Optional[Union[str, bytes]]
    :keyword reply_to: The address of the node to send replies to.
    :paramtype reply_to: Optional[Union[str, bytes]]
    :keyword correlation_id: This is a client-specific id that can be used to mark or identify messages between clients.
    :paramtype correlation_id: Optional[Union[str, bytes]]
    :keyword content_type: The RFC-2046 MIME type for the message's application-data section (body).
    :paramtype content_type: Optional[Union[str, bytes]]
    :keyword content_encoding: The content-encoding property is used as a modifier to the content-type.
    :paramtype content_encoding: Optional[Union[str, bytes]]
    :keyword creation_time: An absolute time when this message was created.
    :paramtype creation_time: Optional[int]
    :keyword absolute_expiry_time: An absolute time when this message is considered to be expired.
    :paramtype absolute_expiry_time: Optional[int]
    :keyword group_id: Identifies the group the message belongs to.
    :paramtype group_id: Optional[Union[str, bytes]]
    :keyword group_sequence: The relative position of this message within its group.
    :paramtype group_sequence: Optional[int]
    :keyword reply_to_group_id: This is a client-specific id that is used so that client can send replies
     to this message to a specific group.
    :paramtype reply_to_group_id: Optional[Union[str, bytes]]
    :ivar message_id: Message-id, if set, uniquely identifies a message within the message system.
     The message producer is usually responsible for setting the message-id in such a way that it
     is assured to be globally unique. A broker MAY discard a message as a duplicate if the value
     of the message-id matches that of a previously received message sent to the same node.
    :vartype message_id: Optional[bytes]
    :ivar user_id: The identity of the user responsible for producing the message. The client sets
     this value, and it MAY be authenticated by intermediaries.
    :vartype user_id: Optional[bytes]
    :ivar to: The to field identifies the node that is the intended destination of the message.
     On any given transfer this might not be the node at the receiving end of the link.
    :vartype to: Optional[bytes]
    :ivar subject: A common field for summary information about the message content and purpose.
    :vartype subject: Optional[bytes]
    :ivar reply_to: The address of the node to send replies to.
    :vartype reply_to: Optional[bytes]
    :ivar correlation_id: his is a client-specific id that can be used to mark or identify messages between clients.
    :vartype correlation_id: Optional[bytes]
    :ivar content_type: The RFC-2046 MIME type for the message's application-data section (body).
    :vartype content_type: Optional[bytes]
    :ivar content_encoding: The content-encoding property is used as a modifier to the content-type.
    :vartype content_encoding: Optional[bytes]
    :ivar creation_time: An absolute time when this message was created.
    :vartype creation_time: Optional[int]
    :ivar absolute_expiry_time: An absolute time when this message is considered to be expired.
    :vartype absolute_expiry_time: Optional[int]
    :ivar group_id: Identifies the group the message belongs to.
    :vartype group_id: Optional[bytes]
    :ivar group_sequence: The relative position of this message within its group.
    :vartype group_sequence: Optional[int]
    :ivar reply_to_group_id: This is a client-specific id that is used so that client can send replies
     to this message to a specific group.
    :vartype reply_to_group_id: Optional[bytes]
    """

    def __init__(self, **kwargs):
        self.message_id = kwargs.get("message_id")
        self.user_id = kwargs.get("user_id")
        self.to = kwargs.get("to")
        self.subject = kwargs.get("subject")
        self.reply_to = kwargs.get("reply_to")
        self.correlation_id = kwargs.get("correlation_id")
        self.content_type = kwargs.get("content_type")
        self.content_encoding = kwargs.get("content_encoding")
        self.creation_time = kwargs.get("creation_time")
        self.absolute_expiry_time = kwargs.get("absolute_expiry_time")
        self.group_id = kwargs.get("group_id")
        self.group_sequence = kwargs.get("group_sequence")
        self.reply_to_group_id = kwargs.get("reply_to_group_id")



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
An example to show receiving events from Event Hub partitions with custom starting position asynchronously.
"""
import os
import asyncio
from azure.eventhub.aio import EventHubConsumerClient, EventHubProducerClient
from azure.eventhub import EventData
from azure.identity.aio import DefaultAzureCredential

FULLY_QUALIFIED_NAMESPACE = os.environ["EVENT_HUB_HOSTNAME"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]


async def on_partition_initialize(partition_context):
    # Put your code here.
    print("Partition: {} has been initialized.".format(partition_context.partition_id))


async def on_partition_close(partition_context, reason):
    # Put your code here.
    print("Partition: {} has been closed, reason for closing: {}.".format(partition_context.partition_id, reason))


async def on_error(partition_context, error):
    # Put your code here. partition_context can be None in the on_error callback.
    if partition_context:
        print(
            "An exception: {} occurred during receiving from Partition: {}.".format(
                partition_context.partition_id, error
            )
        )
    else:
        print("An exception: {} occurred during the load balance process.".format(error))


async def on_event(partition_context, event):
    # Put your code here.
    print("Received event: {} from partition: {}.".format(event.body_as_str(), partition_context.partition_id))
    await partition_context.update_checkpoint(event)


async def main():
    producer_client = EventHubProducerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        eventhub_name=EVENTHUB_NAME,
        credential=DefaultAzureCredential(),
    )

    async with producer_client:
        event_data_batch_to_partition_0 = await producer_client.create_batch(partition_id="0")
        event_data_batch_to_partition_0.add(EventData("First event in partition 0"))
        event_data_batch_to_partition_0.add(EventData("Second event in partition 0"))
        event_data_batch_to_partition_0.add(EventData("Third event in partition 0"))
        event_data_batch_to_partition_0.add(EventData("Forth event in partition 0"))
        await producer_client.send_batch(event_data_batch_to_partition_0)

    consumer_client = EventHubConsumerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        eventhub_name=EVENTHUB_NAME,
        credential=DefaultAzureCredential(),
        consumer_group="$Default",
    )

    partition_0_prop = await consumer_client.get_partition_properties("0")
    partition_0_last_enqueued_sequence_number = partition_0_prop["last_enqueued_sequence_number"]

    # client will receive messages from the position of the third from last event on partition 0.
    starting_position = {
        "0": partition_0_last_enqueued_sequence_number - 3,
    }

    async with consumer_client:
        await consumer_client.receive(
            partition_id="0",
            on_event=on_event,
            on_partition_initialize=on_partition_initialize,
            on_partition_close=on_partition_close,
            on_error=on_error,
            starting_position=starting_position,
        )


if __name__ == "__main__":
    asyncio.run(main())



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
"""
An example to show authentication using credentials defined by azure.identity library asynchronously.

EnvironmentCredential is capable of authenticating as a service principal using a client secret or a certificate, or as
a user with a username and password. Configuration is attempted in this order, using these environment variables:

Service principal with secret:
  - **AZURE_TENANT_ID**: ID of the service principal's tenant. Also called its 'directory' ID.
  - **AZURE_CLIENT_ID**: the service principal's client ID
  - **AZURE_CLIENT_SECRET**: one of the service principal's client secrets

Service principal with certificate:
  - **AZURE_TENANT_ID**: ID of the service principal's tenant. Also called its 'directory' ID.
  - **AZURE_CLIENT_ID**: the service principal's client ID
  - **AZURE_CLIENT_CERTIFICATE_PATH**: path to a PEM-encoded certificate file including the private key. The
    certificate must not be password-protected.

User with username and password:
  - **AZURE_CLIENT_ID**: the application's client ID
  - **AZURE_USERNAME**: a username (usually an email address)
  - **AZURE_PASSWORD**: that user's password
  - **AZURE_TENANT_ID**: (optional) ID of the service principal's tenant. Also called its 'directory' ID.
    If not provided, defaults to the 'organizations' tenant, which supports only Azure Active Directory work or
    school accounts.

Please refer to azure.identity library for detailed information.
"""

import os
import asyncio
from azure.eventhub import EventData
from azure.eventhub.aio import EventHubProducerClient
from azure.identity.aio import EnvironmentCredential


fully_qualified_namespace = os.environ["EVENT_HUB_HOSTNAME"]
eventhub_name = os.environ["EVENT_HUB_NAME"]


async def run():
    credential = EnvironmentCredential()
    async with credential:
        producer = EventHubProducerClient(
            fully_qualified_namespace=fully_qualified_namespace, eventhub_name=eventhub_name, credential=credential
        )

        async with producer:
            event_data_batch = await producer.create_batch()
            while True:
                try:
                    event_data_batch.add(EventData("Message inside EventBatchData"))
                except ValueError:
                    # EventDataBatch object reaches max_size.
                    # New EventDataBatch object can be created here to send more data.
                    break
            await producer.send_batch(event_data_batch)


asyncio.run(run())
print("Finished sending.")



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
"""
An example to show authentication using credentials defined by azure.identity library.

EnvironmentCredential is capable of authenticating as a service principal using a client secret or a certificate, or as
a user with a username and password. Configuration is attempted in this order, using these environment variables:

Service principal with secret:
  - **AZURE_TENANT_ID**: ID of the service principal's tenant. Also called its 'directory' ID.
  - **AZURE_CLIENT_ID**: the service principal's client ID
  - **AZURE_CLIENT_SECRET**: one of the service principal's client secrets

Service principal with certificate:
  - **AZURE_TENANT_ID**: ID of the service principal's tenant. Also called its 'directory' ID.
  - **AZURE_CLIENT_ID**: the service principal's client ID
  - **AZURE_CLIENT_CERTIFICATE_PATH**: path to a PEM-encoded certificate file including the private key. The
    certificate must not be password-protected.

User with username and password:
  - **AZURE_CLIENT_ID**: the application's client ID
  - **AZURE_USERNAME**: a username (usually an email address)
  - **AZURE_PASSWORD**: that user's password
  - **AZURE_TENANT_ID**: (optional) ID of the service principal's tenant. Also called its 'directory' ID.
    If not provided, defaults to the 'organizations' tenant, which supports only Azure Active Directory work or
    school accounts.

Please refer to azure.identity library for detailed information.

This sample also shows the process of utilizing a different credential object, in this case, DefaultAzureCredential,
both to demonstrate the ease of adjusting authentication, and to surface another method for doing so.
"""

import os
from azure.eventhub import EventData, EventHubProducerClient
from azure.identity import EnvironmentCredential


fully_qualified_namespace = os.environ["EVENT_HUB_HOSTNAME"]
eventhub_name = os.environ["EVENT_HUB_NAME"]

credential = EnvironmentCredential()

# Note: One has other options to specify the credential.  For instance, DefaultAzureCredential.
# Default Azure Credentials attempt a chained set of authentication methods, per documentation here: https://github.com/Azure/azure-sdk-for-python/tree/main/sdk/identity/azure-identity
# For example user to be logged in can be specified by the environment variable AZURE_USERNAME, consumed via the ManagedIdentityCredential
# Alternately, one can specify the AZURE_TENANT_ID, AZURE_CLIENT_ID, and AZURE_CLIENT_SECRET to use the EnvironmentCredentialClass.
# The docs above specify all mechanisms which the defaultCredential internally support.
#
# credential = DefaultAzureCredential()

producer = EventHubProducerClient(
    fully_qualified_namespace=fully_qualified_namespace, eventhub_name=eventhub_name, credential=credential
)

with producer:
    event_data_batch = producer.create_batch()
    while True:
        try:
            event_data_batch.add(EventData("Message inside EventBatchData"))
        except ValueError:
            # EventDataBatch object reaches max_size.
            # New EventDataBatch object can be created here to send more data.
            break
    producer.send_batch(event_data_batch)

print("Finished sending.")



from unittest.mock import Mock
from azure.eventhub._pyamqp import Connection


def test_connection_begin_session():
    connection = Connection("sb://fake.host.com")
    connection._process_outgoing_frame = Mock(return_value=None)
    # create session on the Connection
    session = connection.create_session(network_trace=False)
    outgoing_channel = session.channel
    # mock starting a begin session to the server
    session.begin = Mock(return_value=None)
    session.begin()
    # in response from the server we should get back a BEGIN frame
    incoming_channel = 0
    incoming_frame = (1, 0, 0, 0, 0, 0, 0, 0)
    connection.listen = Mock(side_effect=connection._incoming_begin(incoming_channel, incoming_frame))
    connection.listen()
    assert incoming_channel in connection._incoming_endpoints
    assert outgoing_channel in connection._outgoing_endpoints
    assert outgoing_channel == connection._incoming_endpoints[incoming_channel].channel
    assert connection._incoming_endpoints[incoming_channel] == connection._outgoing_endpoints[outgoing_channel]


def test_connection_end_session_on_timeout():
    connection = Connection("sb://fake.host.com")
    connection._process_outgoing_frame = Mock(return_value=None)
    # create session on the Connection
    session = connection.create_session(network_trace=False)
    outgoing_channel = session.channel
    # mock starting a begin session to the server
    session.begin = Mock(return_value=None)
    session.begin()
    # in response from the server we should get back a BEGIN frame
    incoming_channel = 0
    incoming_frame = (1, 0, 0, 0, 0, 0, 0, 0)
    connection.listen = Mock(side_effect=connection._incoming_begin(incoming_channel, incoming_frame))
    connection.listen()
    assert outgoing_channel == connection._incoming_endpoints[incoming_channel].channel
    # typically after some inactivity(60,000 ms) on the connection, the server will force close a connection
    # and will send across an END frame
    connection.listen = Mock(side_effect=connection._incoming_end(incoming_channel, None))
    connection.listen()
    # end points dont have the channels anymore and Session link is gone.
    assert outgoing_channel not in connection._outgoing_endpoints
    assert incoming_channel not in connection._incoming_endpoints



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
An example to show receiving events from Event Hub partitions with custom starting position.
"""
import os
from azure.eventhub import EventHubConsumerClient, EventHubProducerClient, EventData
from azure.identity import DefaultAzureCredential

FULLY_QUALIFIED_NAMESPACE = os.environ["EVENT_HUB_HOSTNAME"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]


def on_partition_initialize(partition_context):
    # Put your code here.
    print("Partition: {} has been initialized.".format(partition_context.partition_id))


def on_partition_close(partition_context, reason):
    # Put your code here.
    print("Partition: {} has been closed, reason for closing: {}.".format(partition_context.partition_id, reason))


def on_error(partition_context, error):
    # Put your code here. partition_context can be None in the on_error callback.
    if partition_context:
        print(
            "An exception: {} occurred during receiving from Partition: {}.".format(
                partition_context.partition_id, error
            )
        )
    else:
        print("An exception: {} occurred during the load balance process.".format(error))


def on_event(partition_context, event):
    # Put your code here.
    print("Received event: {} from partition: {}.".format(event.body_as_str(), partition_context.partition_id))


if __name__ == "__main__":

    producer_client = EventHubProducerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        eventhub_name=EVENTHUB_NAME,
        credential=DefaultAzureCredential(),
    )

    with producer_client:
        event_data_batch_to_partition_0 = producer_client.create_batch(partition_id="0")
        event_data_batch_to_partition_0.add(EventData("First event in partition 0"))
        event_data_batch_to_partition_0.add(EventData("Second event in partition 0"))
        event_data_batch_to_partition_0.add(EventData("Third event in partition 0"))
        event_data_batch_to_partition_0.add(EventData("Forth event in partition 0"))
        producer_client.send_batch(event_data_batch_to_partition_0)

    consumer_client = EventHubConsumerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        consumer_group="$Default",
        eventhub_name=EVENTHUB_NAME,
        credential=DefaultAzureCredential(),
    )

    partition_0_prop = consumer_client.get_partition_properties("0")
    partition_0_last_enqueued_sequence_number = partition_0_prop["last_enqueued_sequence_number"]

    # client will receive messages from the position of the third from last event on partition 0.
    starting_position = {
        "0": partition_0_last_enqueued_sequence_number - 3,
    }

    try:
        with consumer_client:
            consumer_client.receive(
                partition_id="0",
                on_event=on_event,
                on_partition_initialize=on_partition_initialize,
                on_partition_close=on_partition_close,
                on_error=on_error,
                starting_position=starting_position,
            )
    except KeyboardInterrupt:
        print("Stopped receiving.")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from ._connection_async import Connection, ConnectionState
from ._link_async import Link, LinkState
from ..constants import LinkDeliverySettleReason
from ._receiver_async import ReceiverLink
from ._sasl_async import SASLPlainCredential, SASLTransport
from ._sender_async import SenderLink
from ._session_async import Session, SessionState
from ._transport_async import AsyncTransport
from ._client_async import AMQPClientAsync, ReceiveClientAsync, SendClientAsync
from ._authentication_async import SASTokenAuthAsync

__all__ = [
    "Connection",
    "ConnectionState",
    "Link",
    "LinkDeliverySettleReason",
    "LinkState",
    "ReceiverLink",
    "SASLPlainCredential",
    "SASLTransport",
    "SenderLink",
    "Session",
    "SessionState",
    "AsyncTransport",
    "AMQPClientAsync",
    "ReceiveClientAsync",
    "SendClientAsync",
    "SASTokenAuthAsync",
]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

# TODO: fix mypy errors for _code/_definition/__defaults__ (issue #26500)
from typing import Any, Union, Optional
from collections import namedtuple

from enum import Enum
from .constants import SECURE_PORT, FIELD
from .types import AMQPTypes, FieldDefinition


class ErrorCondition(bytes, Enum):
    # Shared error conditions:

    #: An internal error occurred. Operator intervention may be required to resume normaloperation.
    InternalError = b"amqp:internal-error"
    #: A peer attempted to work with a remote entity that does not exist.
    NotFound = b"amqp:not-found"
    #: A peer attempted to work with a remote entity to which it has no access due tosecurity settings.
    UnauthorizedAccess = b"amqp:unauthorized-access"
    #: Data could not be decoded.
    DecodeError = b"amqp:decode-error"
    #: A peer exceeded its resource allocation.
    ResourceLimitExceeded = b"amqp:resource-limit-exceeded"
    #: The peer tried to use a frame in a manner that is inconsistent with the semantics defined in the specification.
    NotAllowed = b"amqp:not-allowed"
    #: An invalid field was passed in a frame body, and the operation could not proceed.
    InvalidField = b"amqp:invalid-field"
    #: The peer tried to use functionality that is not implemented in its partner.
    NotImplemented = b"amqp:not-implemented"
    #: The client attempted to work with a server entity to which it has no access
    #: because another client is working with it.
    ResourceLocked = b"amqp:resource-locked"
    #: The client made a request that was not allowed because some precondition failed.
    PreconditionFailed = b"amqp:precondition-failed"
    #: A server entity the client is working with has been deleted.
    ResourceDeleted = b"amqp:resource-deleted"
    #: The peer sent a frame that is not permitted in the current state of the Session.
    IllegalState = b"amqp:illegal-state"
    #: The peer cannot send a frame because the smallest encoding of the performative with the currently
    #: valid values would be too large to fit within a frame of the agreed maximum frame size.
    FrameSizeTooSmall = b"amqp:frame-size-too-small"

    # Symbols used to indicate connection error conditions:

    #: An operator intervened to close the Connection for some reason. The client may retry at some later date.
    ConnectionCloseForced = b"amqp:connection:forced"
    #: A valid frame header cannot be formed from the incoming byte stream.
    ConnectionFramingError = b"amqp:connection:framing-error"
    #: The container is no longer available on the current connection. The peer should attempt reconnection
    #: to the container using the details provided in the info map.
    ConnectionRedirect = b"amqp:connection:redirect"

    # Symbols used to indicate session error conditions:

    #: The peer violated incoming window for the session.
    SessionWindowViolation = b"amqp:session:window-violation"
    #: Input was received for a link that was detached with an error.
    SessionErrantLink = b"amqp:session:errant-link"
    #: An attach was received using a handle that is already in use for an attached Link.
    SessionHandleInUse = b"amqp:session:handle-in-use"
    #: A frame (other than attach) was received referencing a handle which
    #: is not currently in use of an attached Link.
    SessionUnattachedHandle = b"amqp:session:unattached-handle"

    # Symbols used to indicate link error conditions:

    #: An operator intervened to detach for some reason.
    LinkDetachForced = b"amqp:link:detach-forced"
    #: The peer sent more Message transfers than currently allowed on the link.
    LinkTransferLimitExceeded = b"amqp:link:transfer-limit-exceeded"
    #: The peer sent a larger message than is supported on the link.
    LinkMessageSizeExceeded = b"amqp:link:message-size-exceeded"
    #: The address provided cannot be resolved to a terminus at the current container.
    LinkRedirect = b"amqp:link:redirect"
    #: The link has been attached elsewhere, causing the existing attachment to be forcibly closed.
    LinkStolen = b"amqp:link:stolen"

    # Customized symbols used to indicate client error conditions.
    # TODO: check whether Client/Unknown/Vendor Error are exposed in EH/SB as users might be depending
    #  on the code for error handling
    ClientError = b"amqp:client-error"
    UnknownError = b"amqp:unknown-error"
    VendorError = b"amqp:vendor-error"
    SocketError = b"amqp:socket-error"


class RetryMode(str, Enum):  # pylint: disable=enum-must-inherit-case-insensitive-enum-meta
    EXPONENTIAL = "exponential"
    FIXED = "fixed"


class RetryPolicy:

    no_retry = [
        ErrorCondition.DecodeError,
        ErrorCondition.LinkMessageSizeExceeded,
        ErrorCondition.NotFound,
        ErrorCondition.NotImplemented,
        ErrorCondition.LinkRedirect,
        ErrorCondition.NotAllowed,
        ErrorCondition.UnauthorizedAccess,
        ErrorCondition.LinkStolen,
        ErrorCondition.ResourceLimitExceeded,
        ErrorCondition.ConnectionRedirect,
        ErrorCondition.PreconditionFailed,
        ErrorCondition.InvalidField,
        ErrorCondition.ResourceDeleted,
        ErrorCondition.IllegalState,
        ErrorCondition.FrameSizeTooSmall,
        ErrorCondition.ConnectionFramingError,
        ErrorCondition.SessionUnattachedHandle,
        ErrorCondition.SessionHandleInUse,
        ErrorCondition.SessionErrantLink,
        ErrorCondition.SessionWindowViolation,
    ]

    def __init__(self, **kwargs):
        """
        keyword int retry_total:
        keyword float retry_backoff_factor:
        keyword float retry_backoff_max:
        keyword RetryMode retry_mode:
        keyword list no_retry:
        keyword dict custom_retry_policy:
        """
        self.total_retries = kwargs.pop("retry_total", 3)
        # TODO: A. consider letting retry_backoff_factor be either a float or a callback obj which returns a float
        #  to give more extensibility on customization of retry backoff time, the callback could take the exception
        #  as input.
        self.backoff_factor = kwargs.pop("retry_backoff_factor", 0.8)
        self.backoff_max = kwargs.pop("retry_backoff_max", 120)
        self.retry_mode = kwargs.pop("retry_mode", RetryMode.EXPONENTIAL)
        self.no_retry.extend(kwargs.get("no_retry", []))
        self.custom_condition_backoff = kwargs.pop("custom_condition_backoff", None)
        # TODO: B. As an alternative of option A, we could have a new kwarg serve the goal

    def configure_retries(self, **kwargs):
        return {
            "total": kwargs.pop("retry_total", self.total_retries),
            "backoff": kwargs.pop("retry_backoff_factor", self.backoff_factor),
            "max_backoff": kwargs.pop("retry_backoff_max", self.backoff_max),
            "retry_mode": kwargs.pop("retry_mode", self.retry_mode),
            "history": [],
        }

    def increment(self, settings, error):
        settings["total"] -= 1
        settings["history"].append(error)
        if settings["total"] < 0:
            return False
        return True

    def is_retryable(self, error):
        try:
            if error.condition in self.no_retry:
                return False
        except TypeError:
            pass
        return True

    def get_backoff_time(self, settings, error):
        try:
            return self.custom_condition_backoff[error.condition]
        except (KeyError, TypeError):
            pass

        consecutive_errors_len = len(settings["history"])
        if consecutive_errors_len <= 1:
            return 0

        if self.retry_mode == RetryMode.FIXED:
            backoff_value = settings["backoff"]
        else:
            backoff_value = settings["backoff"] * (2 ** (consecutive_errors_len - 1))
        return min(settings["max_backoff"], backoff_value)


AMQPError = namedtuple("AMQPError", ["condition", "description", "info"], defaults=[None, None])
AMQPError.__new__.__defaults__ = (None,) * len(AMQPError._fields)  # type: ignore
AMQPError._code = 0x0000001D  # type: ignore # pylint: disable=protected-access
AMQPError._definition = (  # type: ignore # pylint: disable=protected-access
    FIELD("condition", AMQPTypes.symbol, True, None, False),
    FIELD("description", AMQPTypes.string, False, None, False),
    FIELD("info", FieldDefinition.fields, False, None, False),
)


class AMQPException(Exception):
    """Base exception for all errors.

    :param bytes condition: The error code.
    :keyword str description: A description of the error.
    :keyword dict info: A dictionary of additional data associated with the error.
    """

    def __init__(self, condition: bytes, **kwargs: Any):
        self.condition: Union[bytes, ErrorCondition] = condition or ErrorCondition.UnknownError
        self.description: Optional[Union[str, bytes]] = kwargs.pop("description", None)
        self.info: Optional[str] = kwargs.get("info", None)
        self.message: Optional[str] = kwargs.get("message", None)
        self.inner_error: Optional[str] = kwargs.get("error", None)
        message = self.message or "Error condition: {}".format(
            str(condition) if isinstance(condition, ErrorCondition) else condition.decode()
        )
        if self.description:
            if isinstance(self.description, bytes):
                message += "\n Error Description: {}".format(self.description.decode())
            else:
                message += "\n Error Description: {}".format(self.description)
        super(AMQPException, self).__init__(message)


class AMQPDecodeError(AMQPException):
    """An error occurred while decoding an incoming frame."""


class AMQPConnectionError(AMQPException):
    """Details of a Connection-level error."""


class AMQPConnectionRedirect(AMQPConnectionError):
    """Details of a Connection-level redirect response.

    The container is no longer available on the current connection.
    The peer should attempt reconnection to the container using the details provided.

    :param bytes condition: The error code.
    :keyword str description: A description of the error.
    :keyword dict info: A dictionary of additional data associated with the error.
    """

    def __init__(self, condition, description=None, info=None):
        self.hostname = info.get(b"hostname", b"").decode("utf-8")
        self.network_host = info.get(b"network-host", b"").decode("utf-8")
        self.port = int(info.get(b"port", SECURE_PORT))
        super(AMQPConnectionRedirect, self).__init__(condition, description=description, info=info)


class AMQPSessionError(AMQPException):
    """Details of a Session-level error.

    :param bytes condition: The error code.
    :keyword str description: A description of the error.
    :keyword dict info: A dictionary of additional data associated with the error.
    """


class AMQPLinkError(AMQPException):
    """Details of a Link-level error.

    :param bytes condition: The error code.
    :keyword str description: A description of the error.
    :keyword dict info: A dictionary of additional data associated with the error.
    """


class AMQPLinkRedirect(AMQPLinkError):
    """Details of a Link-level redirect response.

    The address provided cannot be resolved to a terminus at the current container.
    The supplied information may allow the client to locate and attach to the terminus.

    :param bytes condition: The error code.
    :keyword str description: A description of the error.
    :keyword dict info: A dictionary of additional data associated with the error.
    """

    def __init__(self, condition, description=None, info=None):
        self.hostname = info.get(b"hostname", b"").decode("utf-8")
        self.network_host = info.get(b"network-host", b"").decode("utf-8")
        self.port = int(info.get(b"port", SECURE_PORT))
        self.address = info.get(b"address", b"").decode("utf-8")
        super().__init__(condition, description=description, info=info)


class AuthenticationException(AMQPException):
    """Details of a Authentication error.

    :param bytes condition: The error code.
    :keyword str description: A description of the error.
    :keyword dict info: A dictionary of additional data associated with the error.
    """


class TokenExpired(AuthenticationException):
    """Details of a Token expiration error.

    :param bytes condition: The error code.
    :keyword str description: A description of the error.
    :keyword dict info: A dictionary of additional data associated with the error.
    """


class TokenAuthFailure(AuthenticationException):
    """Failure to authenticate with token."""

    def __init__(self, status_code, status_description, **kwargs):
        encoding = kwargs.get("encoding", "utf-8")
        self.status_code = status_code
        self.status_description = status_description
        message = "CBS Token authentication failed.\nStatus code: {}".format(self.status_code)
        if self.status_description:
            try:
                message += "\nDescription: {}".format(self.status_description.decode(encoding))
            except (TypeError, AttributeError):
                message += "\nDescription: {}".format(self.status_description)
        super(TokenAuthFailure, self).__init__(condition=ErrorCondition.ClientError, message=message)


class MessageException(AMQPException):
    """Details of a Message error.

    :param bytes condition: The error code.
    :keyword str description: A description of the error.
    :keyword dict info: A dictionary of additional data associated with the error.

    """


class MessageSendFailed(MessageException):
    """Details of a Message send failed error.

    :param bytes condition: The error code.
    :keyword str description: A description of the error.
    :keyword dict info: A dictionary of additional data associated with the error.
    """


class ErrorResponse(object):
    """AMQP error object."""

    def __init__(self, **kwargs):
        self.condition = kwargs.get("condition")
        self.description = kwargs.get("description")

        info = kwargs.get("info")
        error_info = kwargs.get("error_info")
        if isinstance(error_info, list) and len(error_info) >= 1:
            if isinstance(error_info[0], list) and len(error_info[0]) >= 1:
                self.condition = error_info[0][0]
                if len(error_info[0]) >= 2:
                    self.description = error_info[0][1]
                if len(error_info[0]) >= 3:
                    info = error_info[0][2]

        self.info = info
        self.error = error_info



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
import time
import sys
import threading

from azure.eventhub import EventData, EventDataBatch
from azure.eventhub.exceptions import (
    ConnectError,
    AuthenticationError,
    EventHubError,
    OperationTimeoutError,
)
from azure.eventhub import (
    EventHubProducerClient,
    EventHubConsumerClient,
    EventHubSharedKeyCredential,
)
from azure.eventhub._client_base import EventHubSASTokenCredential


@pytest.mark.liveTest
def test_send_batch_with_invalid_hostname(invalid_hostname, uamqp_transport, client_args):
    if sys.platform.startswith("darwin"):
        pytest.skip(
            "Skipping on OSX - it keeps reporting 'Unable to set external certificates' " "and blocking other tests"
        )
    client = EventHubProducerClient.from_connection_string(invalid_hostname, uamqp_transport=uamqp_transport, **client_args)
    with client:
        with pytest.raises(ConnectError):
            batch = EventDataBatch()
            batch.add(EventData("test data"))
            client.send_batch(batch)

    # test setting callback
    def on_error(events, pid, err):
        assert len(events) == 1
        assert not pid
        on_error.err = err

    on_error.err = None
    client = EventHubProducerClient.from_connection_string(
        invalid_hostname, on_error=on_error, uamqp_transport=uamqp_transport, **client_args
    )
    with client:
        batch = EventDataBatch()
        batch.add(EventData("test data"))
        client.send_batch(batch)
    assert isinstance(on_error.err, ConnectError)

    on_error.err = None
    client = EventHubProducerClient.from_connection_string(
        invalid_hostname, on_error=on_error, uamqp_transport=uamqp_transport, **client_args
    )
    with client:
        client.send_event(EventData("test data"))
    assert isinstance(on_error.err, ConnectError)


@pytest.mark.liveTest
def test_receive_with_invalid_hostname_sync(invalid_hostname, uamqp_transport, client_args):
    def on_event(partition_context, event):
        pass

    client = EventHubConsumerClient.from_connection_string(
        invalid_hostname, consumer_group="$default", uamqp_transport=uamqp_transport, **client_args
    )
    with client:
        thread = threading.Thread(target=client.receive, args=(on_event,))
        thread.start()
        time.sleep(2)
        assert len(client._event_processors) == 1
    thread.join()


@pytest.mark.liveTest
def test_send_batch_with_invalid_key(invalid_key, uamqp_transport, client_args):
    client = EventHubProducerClient.from_connection_string(invalid_key, uamqp_transport=uamqp_transport, **client_args)
    try:
        with pytest.raises(ConnectError):
            batch = EventDataBatch()
            batch.add(EventData("test data"))
            client.send_batch(batch)
    finally:
        client.close()


@pytest.mark.liveTest
def test_send_batch_to_invalid_partitions(auth_credentials, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential = auth_credentials
    partitions = ["XYZ", "-1", "1000", "-"]
    for p in partitions:
        client = EventHubProducerClient(
            fully_qualified_namespace=fully_qualified_namespace,
            eventhub_name=eventhub_name,
            credential=credential(),
            uamqp_transport=uamqp_transport,
            **client_args,
        )
        try:
            with pytest.raises(ConnectError):
                batch = client.create_batch(partition_id=p)
                batch.add(EventData("test data"))
                client.send_batch(batch)
        finally:
            client.close()


@pytest.mark.liveTest
def test_send_batch_too_large_message(auth_credentials, uamqp_transport, client_args):
    if sys.platform.startswith("darwin"):
        pytest.skip("Skipping on OSX - open issue regarding message size")
    fully_qualified_namespace, eventhub_name, credential = auth_credentials
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    try:
        data = EventData(b"A" * 1100000)
        batch = client.create_batch()
        with pytest.raises(ValueError):
            batch.add(data)
    finally:
        client.close()


@pytest.mark.liveTest
def test_send_batch_null_body(auth_credentials, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential = auth_credentials
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    try:
        with pytest.raises(ValueError):
            data = EventData(None)
            batch = client.create_batch()
            batch.add(data)
            client.send_batch(batch)
    finally:
        client.close()


@pytest.mark.liveTest
def test_create_batch_with_invalid_hostname_sync(invalid_hostname, uamqp_transport, client_args):
    if sys.platform.startswith("darwin"):
        pytest.skip(
            "Skipping on OSX - it keeps reporting 'Unable to set external certificates' " "and blocking other tests"
        )
    client = EventHubProducerClient.from_connection_string(invalid_hostname, uamqp_transport=uamqp_transport, **client_args)
    with client:
        with pytest.raises(ConnectError):
            client.create_batch(max_size_in_bytes=300)


@pytest.mark.liveTest
def test_create_batch_with_too_large_size_sync(auth_credentials, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential = auth_credentials
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    with client:
        with pytest.raises(ValueError):
            client.create_batch(max_size_in_bytes=5 * 1024 * 1024)


@pytest.mark.liveTest
def test_invalid_proxy_server(auth_credentials, uamqp_transport, client_args):
    if sys.platform.startswith("darwin") and uamqp_transport:
        pytest.skip("Skipping on OSX - running forever and blocking other tests")
    fully_qualified_namespace, eventhub_name, credential = auth_credentials
    HTTP_PROXY = {
        "proxy_hostname": "fakeproxy",  # proxy hostname.
        "proxy_port": 3128,  # proxy port.
    }

    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        http_proxy=HTTP_PROXY,
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    with client:
        with pytest.raises(EventHubError):
            batch = client.create_batch()


@pytest.mark.liveTest
def test_client_send_timeout(auth_credential_receivers, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential, receivers = auth_credential_receivers

    def on_success(events, pid):
        pass

    def on_error(events, pid, err):
        pass

    producer = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        uamqp_transport=uamqp_transport,
        **client_args,
    )

    with producer:
        with pytest.raises(OperationTimeoutError):
            producer.send_batch([EventData(b"Data")], timeout=-1)

        with pytest.raises(OperationTimeoutError):
            producer.send_event(EventData(b"Data"), timeout=-1)

    producer = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        buffered_mode=True,
        on_success=on_success,
        on_error=on_error,
        uamqp_transport=uamqp_transport,
        **client_args,
    )

    with producer:
        with pytest.raises(OperationTimeoutError):
            producer.send_batch([EventData(b"Data")], timeout=-1)

        with pytest.raises(OperationTimeoutError):
            producer.send_event(EventData(b"Data"), timeout=-1)


@pytest.mark.liveTest
# Not recording since proxy requires a specific ssl_context, which causes the test to fail when connection_verify is passed
@pytest.mark.no_amqpproxy
def test_client_invalid_credential(live_eventhub, get_credential, uamqp_transport, client_args):

    def on_event(partition_context, event):
        pass

    def on_error(partition_context, error):
        on_error.err = error

    azure_credential = get_credential()
    # Skipping on OSX - it's raising a ConnectionLostError
    if not sys.platform.startswith("darwin"):
        producer_client = EventHubProducerClient(
            fully_qualified_namespace="fakeeventhub.servicebus.windows.net",
            eventhub_name=live_eventhub["event_hub"],
            credential=azure_credential,
            user_agent="customized information",
            retry_total=1,
            retry_mode="exponential",
            retry_backoff=0.02,
            uamqp_transport=uamqp_transport,
            **client_args,
        )
        consumer_client = EventHubConsumerClient(
            fully_qualified_namespace="fakeeventhub.servicebus.windows.net",
            eventhub_name=live_eventhub["event_hub"],
            credential=azure_credential,
            user_agent="customized information",
            consumer_group="$Default",
            retry_total=1,
            retry_mode="exponential",
            retry_backoff=0.02,
            uamqp_transport=uamqp_transport,
            **client_args,
        )
        with producer_client:
            with pytest.raises(ConnectError):
                producer_client.create_batch(partition_id="0")

        on_error.err = None
        with consumer_client:
            thread = threading.Thread(
                target=consumer_client.receive,
                args=(on_event,),
                kwargs={"starting_position": "-1", "on_error": on_error},
            )
            thread.daemon = True
            thread.start()
            time.sleep(15)
        thread.join()
        assert isinstance(on_error.err, ConnectError)

    producer_client = EventHubProducerClient(
        fully_qualified_namespace=live_eventhub["hostname"],
        eventhub_name="fakehub",
        credential=azure_credential,
        uamqp_transport=uamqp_transport,
        **client_args,
    )

    consumer_client = EventHubConsumerClient(
        fully_qualified_namespace=live_eventhub["hostname"],
        eventhub_name="fakehub",
        credential=azure_credential,
        consumer_group="$Default",
        retry_total=0,
        uamqp_transport=uamqp_transport,
        **client_args,
    )

    with producer_client:
        with pytest.raises(ConnectError):
            producer_client.create_batch(partition_id="0")

    on_error.err = None
    with consumer_client:
        thread = threading.Thread(
            target=consumer_client.receive,
            args=(on_event,),
            kwargs={"starting_position": "-1", "on_error": on_error},
        )
        thread.daemon = True
        thread.start()
        time.sleep(15)
    thread.join()
    assert isinstance(on_error.err, AuthenticationError)

    credential = EventHubSharedKeyCredential(live_eventhub["key_name"], live_eventhub["access_key"])
    auth_uri = "sb://{}/{}".format(live_eventhub["hostname"], live_eventhub["event_hub"])
    token = credential.get_token(auth_uri).token
    producer_client = EventHubProducerClient(
        fully_qualified_namespace=live_eventhub["hostname"],
        eventhub_name=live_eventhub["event_hub"],
        credential=EventHubSASTokenCredential(token, time.time() + 5),
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    time.sleep(10)
    # expired credential
    with producer_client:
        with pytest.raises(AuthenticationError):
            producer_client.create_batch(partition_id="0")

    consumer_client = EventHubConsumerClient(
        fully_qualified_namespace=live_eventhub["hostname"],
        eventhub_name=live_eventhub["event_hub"],
        credential=EventHubSASTokenCredential(token, time.time() + 7),
        consumer_group="$Default",
        retry_total=0,
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    on_error.err = None
    with consumer_client:
        thread = threading.Thread(
            target=consumer_client.receive,
            args=(on_event,),
            kwargs={"starting_position": "-1", "on_error": on_error},
        )
        thread.daemon = True
        thread.start()
        time.sleep(15)
    thread.join()

    assert isinstance(on_error.err, AuthenticationError)

    credential = EventHubSharedKeyCredential("fakekey", live_eventhub["access_key"])
    producer_client = EventHubProducerClient(
        fully_qualified_namespace=live_eventhub["hostname"],
        eventhub_name=live_eventhub["event_hub"],
        credential=credential,
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    with producer_client:
        with pytest.raises(AuthenticationError):
            producer_client.create_batch(partition_id="0")

    consumer_client = EventHubConsumerClient(
        fully_qualified_namespace=live_eventhub["hostname"],
        eventhub_name=live_eventhub["event_hub"],
        credential=credential,
        consumer_group="$Default",
        retry_total=0,
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    on_error.err = None
    with consumer_client:
        thread = threading.Thread(
            target=consumer_client.receive,
            args=(on_event,),
            kwargs={"starting_position": "-1", "on_error": on_error},
        )
        thread.daemon = True
        thread.start()
        time.sleep(15)
    thread.join()
    assert isinstance(on_error.err, AuthenticationError)

    credential = EventHubSharedKeyCredential(live_eventhub["key_name"], "fakekey")
    producer_client = EventHubProducerClient(
        fully_qualified_namespace=live_eventhub["hostname"],
        eventhub_name=live_eventhub["event_hub"],
        credential=credential,
        uamqp_transport=uamqp_transport,
        **client_args,
    )

    with producer_client:
        errors = AuthenticationError
        # TODO: flaky TimeoutError during connect for China region
        if "servicebus.windows.net" not in live_eventhub["hostname"]:
            errors = (AuthenticationError, ConnectError)
        with pytest.raises(errors):
            producer_client.create_batch(partition_id="0")

    consumer_client = EventHubConsumerClient(
        fully_qualified_namespace=live_eventhub["hostname"],
        eventhub_name=live_eventhub["event_hub"],
        credential=credential,
        consumer_group="$Default",
        retry_total=0,
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    on_error.err = None
    with consumer_client:
        thread = threading.Thread(
            target=consumer_client.receive,
            args=(on_event,),
            kwargs={"starting_position": "-1", "on_error": on_error},
        )
        thread.daemon = True
        thread.start()
        time.sleep(15)
    thread.join()
    assert isinstance(on_error.err, AuthenticationError)

    producer_client = EventHubProducerClient(
        fully_qualified_namespace=live_eventhub["hostname"],
        eventhub_name=live_eventhub["event_hub"],
        credential=azure_credential,
        connection_verify="cacert.pem",
        uamqp_transport=uamqp_transport,
    )

    # TODO: this seems like a bug from uamqp, should be ConnectError?
    with producer_client:
        with pytest.raises(EventHubError):
            producer_client.create_batch(partition_id="0")

    consumer_client = EventHubConsumerClient(
        fully_qualified_namespace=live_eventhub["hostname"],
        eventhub_name=live_eventhub["event_hub"],
        consumer_group="$Default",
        credential=azure_credential,
        retry_total=0,
        connection_verify="cacert.pem",
        uamqp_transport=uamqp_transport,
    )
    with consumer_client:
        thread = threading.Thread(
            target=consumer_client.receive,
            args=(on_event,),
            kwargs={"starting_position": "-1", "on_error": on_error},
        )
        thread.daemon = True
        thread.start()
        time.sleep(15)
    thread.join()

    # TODO: this seems like a bug from uamqp, should be ConnectError?
    assert isinstance(on_error.err, FileNotFoundError)

    # Skipping on OSX - it's raising a ConnectionLostError
    if not sys.platform.startswith("darwin"):
        producer_client = EventHubProducerClient(
            fully_qualified_namespace=live_eventhub["hostname"],
            eventhub_name=live_eventhub["event_hub"],
            credential=azure_credential,
            custom_endpoint_address="fakeaddr",
            uamqp_transport=uamqp_transport,
            **client_args,
        )

        with producer_client:
            with pytest.raises(AuthenticationError):
                producer_client.create_batch(partition_id="0")

        consumer_client = EventHubConsumerClient(
            fully_qualified_namespace=live_eventhub["hostname"],
            eventhub_name=live_eventhub["event_hub"],
            consumer_group="$Default",
            credential=azure_credential,
            retry_total=0,
            custom_endpoint_address="fakeaddr",
            uamqp_transport=uamqp_transport,
            **client_args,
        )
        with consumer_client:
            thread = threading.Thread(
                target=consumer_client.receive,
                args=(on_event,),
                kwargs={"starting_position": "-1", "on_error": on_error},
            )
            thread.daemon = True
            thread.start()
            time.sleep(15)
        thread.join()

        assert isinstance(on_error.err, AuthenticationError)



"""Platform compatibility."""

# pylint: skip-file

from typing import Tuple, cast
import platform
import re
import struct
import sys

# Jython does not have this attribute
try:
    from socket import SOL_TCP
except ImportError:  # pragma: no cover
    from socket import IPPROTO_TCP as SOL_TCP  # noqa


RE_NUM = re.compile(r"(\d+).+")


def _linux_version_to_tuple(s):
    # type: (str) -> Tuple[int, int, int]
    return cast(Tuple[int, int, int], tuple(map(_versionatom, s.split(".")[:3])))


def _versionatom(s):
    # type: (str) -> int
    if s.isdigit():
        return int(s)
    match = RE_NUM.match(s)
    return int(match.groups()[0]) if match else 0


# available socket options for TCP level
KNOWN_TCP_OPTS = {
    "TCP_CORK",
    "TCP_DEFER_ACCEPT",
    "TCP_KEEPCNT",
    "TCP_KEEPIDLE",
    "TCP_KEEPINTVL",
    "TCP_LINGER2",
    "TCP_MAXSEG",
    "TCP_NODELAY",
    "TCP_QUICKACK",
    "TCP_SYNCNT",
    "TCP_USER_TIMEOUT",
    "TCP_WINDOW_CLAMP",
}

LINUX_VERSION = None
if sys.platform.startswith("linux"):
    LINUX_VERSION = _linux_version_to_tuple(platform.release())
    if LINUX_VERSION < (2, 6, 37):
        KNOWN_TCP_OPTS.remove("TCP_USER_TIMEOUT")

    # Windows Subsystem for Linux is an edge-case: the Python socket library
    # returns most TCP_* enums, but they aren't actually supported
    if platform.release().endswith("Microsoft"):
        KNOWN_TCP_OPTS = {"TCP_NODELAY", "TCP_KEEPIDLE", "TCP_KEEPINTVL", "TCP_KEEPCNT"}

elif sys.platform.startswith("darwin"):
    KNOWN_TCP_OPTS.remove("TCP_USER_TIMEOUT")

elif "bsd" in sys.platform:
    KNOWN_TCP_OPTS.remove("TCP_USER_TIMEOUT")

# According to MSDN Windows platforms support getsockopt(TCP_MAXSSEG) but not
# setsockopt(TCP_MAXSEG) on IPPROTO_TCP sockets.
elif sys.platform.startswith("win"):
    KNOWN_TCP_OPTS = {"TCP_NODELAY"}

elif sys.platform.startswith("cygwin"):
    KNOWN_TCP_OPTS = {"TCP_NODELAY"}

# illumos does not allow to set the TCP_MAXSEG socket option,
# even if the Oracle documentation says otherwise.
elif sys.platform.startswith("sunos"):
    KNOWN_TCP_OPTS.remove("TCP_MAXSEG")

# aix does not allow to set the TCP_MAXSEG
# or the TCP_USER_TIMEOUT socket options.
elif sys.platform.startswith("aix"):
    KNOWN_TCP_OPTS.remove("TCP_MAXSEG")
    KNOWN_TCP_OPTS.remove("TCP_USER_TIMEOUT")

pack = struct.pack
pack_into = struct.pack_into
unpack = struct.unpack
unpack_from = struct.unpack_from

__all__ = [
    "LINUX_VERSION",
    "SOL_TCP",
    "KNOWN_TCP_OPTS",
    "pack",
    "pack_into",
    "unpack",
    "unpack_from",
]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from ._transport_async import AsyncTransport, WebSocketTransportAsync
from ..constants import SASLCode, SASL_HEADER_FRAME, WEBSOCKET_PORT
from .._transport import AMQPS_PORT
from ..performatives import SASLInit


_SASL_FRAME_TYPE = b"\x01"


# TODO: do we need it here? it's a duplicate of the sync version
class SASLPlainCredential(object):
    """PLAIN SASL authentication mechanism.
    See https://tools.ietf.org/html/rfc4616 for details
    """

    mechanism = b"PLAIN"

    def __init__(self, authcid, passwd, authzid=None):
        self.authcid = authcid
        self.passwd = passwd
        self.authzid = authzid

    def start(self):
        if self.authzid:
            login_response = self.authzid.encode("utf-8")
        else:
            login_response = b""
        login_response += b"\0"
        login_response += self.authcid.encode("utf-8")
        login_response += b"\0"
        login_response += self.passwd.encode("utf-8")
        return login_response


# TODO: do we need it here? it's a duplicate of the sync version
class SASLAnonymousCredential(object):
    """ANONYMOUS SASL authentication mechanism.
    See https://tools.ietf.org/html/rfc4505 for details
    """

    mechanism = b"ANONYMOUS"

    def start(self):
        return b""


# TODO: do we need it here? it's a duplicate of the sync version
class SASLExternalCredential(object):
    """EXTERNAL SASL mechanism.
    Enables external authentication, i.e. not handled through this protocol.
    Only passes 'EXTERNAL' as authentication mechanism, but no further
    authentication data.
    """

    mechanism = b"EXTERNAL"

    def start(self):
        return b""


class SASLTransportMixinAsync:
    async def _negotiate(self):
        await self.write(SASL_HEADER_FRAME)
        _, returned_header = await self.receive_frame()
        if returned_header[1] != SASL_HEADER_FRAME:
            raise ValueError(
                f"""Mismatching AMQP header protocol. Expected: {SASL_HEADER_FRAME!r},"""
                """received: {returned_header[1]!r}"""
            )

        _, supported_mechanisms = await self.receive_frame(verify_frame_type=1)
        if self.credential.mechanism not in supported_mechanisms[1][0]:  # sasl_server_mechanisms
            raise ValueError("Unsupported SASL credential type: {}".format(self.credential.mechanism))
        sasl_init = SASLInit(
            mechanism=self.credential.mechanism,
            initial_response=self.credential.start(),
            hostname=self.host,
        )
        await self.send_frame(0, sasl_init, frame_type=_SASL_FRAME_TYPE)

        _, next_frame = await self.receive_frame(verify_frame_type=1)
        frame_type, fields = next_frame
        if frame_type != 0x00000044:  # SASLOutcome
            raise NotImplementedError("Unsupported SASL challenge")
        if fields[0] == SASLCode.Ok:  # code
            return
        raise ValueError("SASL negotiation failed.\nOutcome: {}\nDetails: {}".format(*fields))


class SASLTransport(AsyncTransport, SASLTransportMixinAsync):
    def __init__(
        self,
        host,
        credential,
        *,
        port=AMQPS_PORT,
        socket_timeout=None,
        ssl_opts=None,
        **kwargs,
    ):
        self.credential = credential
        ssl_opts = ssl_opts or True
        super(SASLTransport, self).__init__(
            host,
            port=port,
            socket_timeout=socket_timeout,
            ssl_opts=ssl_opts,
            **kwargs,
        )

    async def negotiate(self):
        await self._negotiate()


class SASLWithWebSocket(WebSocketTransportAsync, SASLTransportMixinAsync):
    def __init__(
        self,
        host,
        credential,
        *,
        port=WEBSOCKET_PORT,
        socket_timeout=None,
        ssl_opts=None,
        **kwargs,
    ):
        self.credential = credential
        ssl_opts = ssl_opts or True
        super().__init__(
            host,
            port=port,
            socket_timeout=socket_timeout,
            ssl_opts=ssl_opts,
            **kwargs,
        )

    async def negotiate(self):
        await self._negotiate()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from __future__ import annotations

import logging
import uuid
import time
import functools
import collections
from typing import Any, Dict, Tuple, List, Optional, TYPE_CHECKING, cast, Union
from datetime import timedelta
from urllib.parse import urlparse
from typing_extensions import TypeAlias

from azure.core.credentials import (
    AccessToken,
    AzureSasCredential,
    AzureNamedKeyCredential,
)
from azure.core.utils import parse_connection_string as core_parse_connection_string
from azure.core.pipeline.policies import RetryMode

try:
    from ._transport._uamqp_transport import UamqpTransport
except ImportError:
    UamqpTransport = None  # type: ignore
from ._transport._pyamqp_transport import PyamqpTransport
from .exceptions import ClientClosedError
from ._configuration import Configuration
from ._utils import utc_from_timestamp, parse_sas_credential
from ._pyamqp.utils import generate_sas_token
from ._connection_manager import get_connection_manager
from ._constants import (
    CONTAINER_PREFIX,
    JWT_TOKEN_SCOPE,
    READ_OPERATION,
    MGMT_STATUS_CODE,
    MGMT_STATUS_DESC,
    MGMT_OPERATION,
    MGMT_PARTITION_OPERATION,
)

if TYPE_CHECKING:
    from azure.core.credentials import TokenCredential

    CredentialTypes: TypeAlias = Union[
        AzureSasCredential,
        AzureNamedKeyCredential,
        "EventHubSharedKeyCredential",
        TokenCredential,
    ]
    from ._consumer_client import EventHubConsumerClient
    from ._producer_client import EventHubProducerClient
    from ._transport._base import AmqpTransport

    try:
        from uamqp import Message as uamqp_Message
        from uamqp.authentication import JWTTokenAuth as uamqp_JWTTokenAuth
        from uamqp import ReceiveClient as uamqp_AMQPRecieveClient
        from uamqp import SendClient as uamqp_AMQPSendClient
    except ImportError:
        pass
    from ._pyamqp.message import Message
    from ._pyamqp.authentication import JWTTokenAuth
    from ._pyamqp import ReceiveClient as pyamqp_AMQPRecieveClient
    from ._pyamqp import SendClient as pyamqp_AMQPSendClient

_LOGGER = logging.getLogger(__name__)
_Address = collections.namedtuple("_Address", "hostname path")


def _parse_conn_str(
    conn_str: str,  # pylint:disable=unused-argument
    *,
    eventhub_name: Optional[str] = None,
    check_case: bool = False,
    **kwargs: Any,
) -> Tuple[str, Optional[str], Optional[str], str, Optional[str], Optional[int], bool]:
    endpoint = None
    shared_access_key_name = None
    shared_access_key = None
    entity_path: Optional[str] = None
    shared_access_signature: Optional[str] = None
    shared_access_signature_expiry = None
    use_emulator: Optional[str] = None
    conn_settings = core_parse_connection_string(conn_str, case_sensitive_keys=check_case)
    if check_case:
        shared_access_key = conn_settings.get("SharedAccessKey")
        shared_access_key_name = conn_settings.get("SharedAccessKeyName")
        endpoint = conn_settings.get("Endpoint")
        entity_path = conn_settings.get("EntityPath")
        # non case sensitive check when parsing connection string for internal use
        for key, value in conn_settings.items():
            # only sas check is non case sensitive for both conn str properties and internal use
            if key.lower() == "sharedaccesssignature":
                shared_access_signature = value
        use_emulator = conn_settings.get("UseDevelopmentEmulator")

    if not check_case:
        endpoint = conn_settings.get("endpoint") or conn_settings.get("hostname")
        if endpoint:
            endpoint = endpoint.rstrip("/")
        shared_access_key_name = conn_settings.get("sharedaccesskeyname")
        shared_access_key = conn_settings.get("sharedaccesskey")
        entity_path = conn_settings.get("entitypath")
        shared_access_signature = conn_settings.get("sharedaccesssignature")
        use_emulator = conn_settings.get("usedevelopmentemulator")

    if shared_access_signature:
        try:
            # Expiry can be stored in the "se=<timestamp>" clause of the token. ('&'-separated key-value pairs)
            shared_access_signature_expiry = int(shared_access_signature.split("se=")[1].split("&")[0])  # type: ignore
        except (
            IndexError,
            TypeError,
            ValueError,
        ):  # Fallback since technically expiry is optional.
            # An arbitrary, absurdly large number, since you can't renew.
            shared_access_signature_expiry = int(time.time() * 2)

    entity = cast(str, eventhub_name or entity_path)

    # check that endpoint is valid
    if not endpoint:
        raise ValueError("Connection string is either blank or malformed.")

    parsed = urlparse(endpoint)
    if not parsed.netloc:
        raise ValueError("Invalid Endpoint on the Connection String.")
    host = cast(str, parsed.netloc.strip())

    emulator = use_emulator == "true"

    if any([shared_access_key, shared_access_key_name]) and not all([shared_access_key, shared_access_key_name]):
        raise ValueError(
            "Invalid connection string. Should be in the format: "
            "Endpoint=sb://<FQDN>/;SharedAccessKeyName=<KeyName>;SharedAccessKey=<KeyValue>"
        )
    # Only connection string parser should check that only one of sas and shared access
    # key exists. For backwards compatibility, client construction should not have this check.
    if check_case and shared_access_signature and shared_access_key:
        raise ValueError("Only one of the SharedAccessKey or SharedAccessSignature must be present.")
    if not shared_access_signature and not shared_access_key:
        raise ValueError("At least one of the SharedAccessKey or SharedAccessSignature must be present.")

    return (
        host,
        str(shared_access_key_name) if shared_access_key_name else None,
        str(shared_access_key) if shared_access_key else None,
        entity,
        str(shared_access_signature) if shared_access_signature else None,
        shared_access_signature_expiry,
        emulator,
    )


def _generate_sas_token(uri: str, policy: str, key: str, expiry: Optional[timedelta] = None) -> AccessToken:
    """Create a shared access signature token as a string literal.

    :param str uri: The resource URI.
    :param str policy: The name of the shared access policy.
    :param str key: The shared access key.
    :param datetime.timedelta or None expiry: The time period that the signature is valid for. Default is 1 hour.
    :return: An AccessToken.
    :rtype: ~azure.core.credentials.AccessToken
    """
    if not expiry:
        expiry = timedelta(hours=1)  # Default to 1 hour.

    abs_expiry = int(time.time()) + expiry.seconds

    token = generate_sas_token(uri, policy, key, abs_expiry)
    return AccessToken(token=token, expires_on=abs_expiry)


def _build_uri(address: str, entity: str) -> str:
    parsed = urlparse(address)
    if parsed.path:
        return address
    if not entity:
        raise ValueError("No EventHub specified")
    address += "/" + str(entity)
    return address


def _get_backoff_time(retry_mode, backoff_factor, backoff_max, retried_times):
    if retry_mode == RetryMode.Fixed:
        backoff_value = backoff_factor
    else:
        backoff_value = backoff_factor * (2**retried_times)
    return min(backoff_max, backoff_value)


class EventHubSharedKeyCredential:
    """The shared access key credential used for authentication.

    :param str policy: The name of the shared access policy.
    :param str key: The shared access key.
    """

    def __init__(self, policy: str, key: str) -> None:
        self.policy = policy
        self.key = key
        self.token_type = b"servicebus.windows.net:sastoken"

    def get_token(self, *scopes: str, **kwargs: Any) -> AccessToken:  # pylint:disable=unused-argument
        if not scopes:
            raise ValueError("No token scope provided.")
        return _generate_sas_token(scopes[0], self.policy, self.key)


class EventhubAzureNamedKeyTokenCredential:
    """The named key credential used for authentication.

    :param credential: The AzureNamedKeyCredential that should be used.
    :type credential: ~azure.core.credentials.AzureNamedKeyCredential
    """

    def __init__(self, azure_named_key_credential: AzureNamedKeyCredential) -> None:
        self._credential = azure_named_key_credential
        self.token_type = b"servicebus.windows.net:sastoken"

    def get_token(self, *scopes: str, **kwargs: Any) -> AccessToken:  # pylint:disable=unused-argument
        if not scopes:
            raise ValueError("No token scope provided.")
        name, key = self._credential.named_key
        return _generate_sas_token(scopes[0], name, key)


class EventHubSASTokenCredential:
    """The shared access token credential used for authentication.

    :param str token: The shared access token string
    :param int expiry: The epoch timestamp
    """

    def __init__(self, token: str, expiry: int) -> None:
        """
        :param str token: The shared access token string
        :param float expiry: The epoch timestamp
        """
        self.token = token
        self.expiry = expiry
        self.token_type = b"servicebus.windows.net:sastoken"

    def get_token(self, *scopes: str, **kwargs: Any) -> AccessToken:  # pylint:disable=unused-argument
        """
        This method is automatically called when token is about to expire.

        :param str scopes: The list of scopes for which the token has access.
        :return: The access token.
        :rtype: ~azure.core.credentials.AccessToken
        """
        return AccessToken(self.token, self.expiry)


class EventhubAzureSasTokenCredential:
    """The shared access token credential used for authentication
    when AzureSasCredential is provided.

    :param azure_sas_credential: The credential to be used for authentication.
    :type azure_sas_credential: ~azure.core.credentials.AzureSasCredential
    """

    def __init__(self, azure_sas_credential: AzureSasCredential) -> None:
        """The shared access token credential used for authentication
         when AzureSasCredential is provided.

        :param azure_sas_credential: The credential to be used for authentication.
        :type azure_sas_credential: ~azure.core.credentials.AzureSasCredential
        """
        self._credential = azure_sas_credential
        self.token_type = b"servicebus.windows.net:sastoken"

    def get_token(self, *scopes: str, **kwargs: Any) -> AccessToken:  # pylint:disable=unused-argument
        """
        This method is automatically called when token is about to expire.

        :param str scopes: The scopes for the token request.
        :return: The access token.
        :rtype: ~azure.core.credentials.AccessToken
        """
        signature, expiry = parse_sas_credential(self._credential)
        return AccessToken(signature, cast(int, expiry))


class ClientBase:  # pylint:disable=too-many-instance-attributes
    def __init__(
        self,
        fully_qualified_namespace: str,
        eventhub_name: str,
        credential: "CredentialTypes",
        **kwargs: Any,
    ) -> None:
        uamqp_transport = kwargs.pop("uamqp_transport", False)
        if uamqp_transport and UamqpTransport is None:
            raise ValueError("To use the uAMQP transport, please install `uamqp>=1.6.0,<2.0.0`.")
        self._amqp_transport = kwargs.pop("amqp_transport", UamqpTransport if uamqp_transport else PyamqpTransport)

        self.eventhub_name = eventhub_name
        if not eventhub_name:
            raise ValueError("The eventhub name can not be None or empty.")
        path = "/" + eventhub_name if eventhub_name else ""
        self._address = _Address(hostname=fully_qualified_namespace, path=path)
        self._container_id = CONTAINER_PREFIX + str(uuid.uuid4())[:8]
        if isinstance(credential, AzureSasCredential):
            self._credential = EventhubAzureSasTokenCredential(credential)
        elif isinstance(credential, AzureNamedKeyCredential):
            self._credential = EventhubAzureNamedKeyTokenCredential(credential)  # type: ignore
        else:
            self._credential = credential  # type: ignore
        self._auto_reconnect = kwargs.get("auto_reconnect", True)
        self._auth_uri: str
        self._eventhub_auth_uri = f"sb://{self._address.hostname}{self._address.path}"
        self._config = Configuration(
            amqp_transport=self._amqp_transport,
            hostname=self._address.hostname,
            **kwargs,
        )
        self._debug = self._config.network_tracing
        kwargs["custom_endpoint_address"] = self._config.custom_endpoint_address
        self._conn_manager = get_connection_manager(amqp_transport=self._amqp_transport, **kwargs)
        self._idle_timeout = kwargs.get("idle_timeout", None)

    @staticmethod
    def _from_connection_string(conn_str: str, **kwargs: Any) -> Dict[str, Any]:
        host, policy, key, entity, token, token_expiry, emulator = _parse_conn_str(conn_str, **kwargs)

        kwargs["fully_qualified_namespace"] = host
        kwargs["eventhub_name"] = entity
        # Check if emulator is in use, unset tls if it is
        if emulator:
            kwargs["use_tls"] = False
        if token and token_expiry:
            kwargs["credential"] = EventHubSASTokenCredential(token, token_expiry)
        elif policy and key:
            kwargs["credential"] = EventHubSharedKeyCredential(policy, key)
        return kwargs

    def _create_auth(self, *, auth_uri: Optional[str] = None) -> Union["uamqp_JWTTokenAuth", JWTTokenAuth]:
        """
        Create an ~uamqp.authentication.SASTokenAuth instance
         to authenticate the session.

        :keyword auth_uri: The URI to authenticate with.
        :paramtype auth_uri: str or None

        :return: The auth for the session.
        :rtype: JWTTokenAuth or uamqp_JWTTokenAuth
        """
        # if auth_uri is not provided, use the default hub one
        entity_auth_uri = auth_uri if auth_uri else self._eventhub_auth_uri

        try:
            # ignore mypy's warning because token_type is Optional
            token_type = self._credential.token_type  # type: ignore
        except AttributeError:
            token_type = b"jwt"
        if token_type == b"servicebus.windows.net:sastoken":
            return self._amqp_transport.create_token_auth(
                entity_auth_uri,
                functools.partial(self._credential.get_token, entity_auth_uri),
                token_type=token_type,
                config=self._config,
                update_token=True,
            )
        return self._amqp_transport.create_token_auth(
            entity_auth_uri,
            functools.partial(self._credential.get_token, JWT_TOKEN_SCOPE),
            token_type=token_type,
            config=self._config,
            update_token=False,
        )

    def _close_connection(self) -> None:
        self._conn_manager.reset_connection_if_broken()

    def _backoff(
        self,
        retried_times: int,
        last_exception: Exception,
        timeout_time: Optional[float] = None,
        entity_name: Optional[str] = None,
    ) -> None:
        entity_name = entity_name or self._container_id
        backoff = _get_backoff_time(
            self._config.retry_mode,
            self._config.backoff_factor,
            self._config.backoff_max,
            retried_times,
        )
        if backoff <= self._config.backoff_max and (
            timeout_time is None or time.time() + backoff <= timeout_time
        ):
            time.sleep(backoff)
            _LOGGER.info(
                "%r has an exception (%r). Retrying...",
                format(entity_name),
                last_exception,
            )
        else:
            _LOGGER.info(
                "%r operation has timed out. Last exception before timeout is (%r)",
                entity_name,
                last_exception,
            )
            raise last_exception

    def _management_request(  # pylint:disable=inconsistent-return-statements
        self, mgmt_msg: Union[uamqp_Message, Message], op_type: bytes
    ) -> Any:
        retried_times = 0
        last_exception = None
        while retried_times <= self._config.max_retries:
            mgmt_auth = self._create_auth()
            mgmt_client = self._amqp_transport.create_mgmt_client(
                self._address, mgmt_auth=mgmt_auth, config=self._config
            )
            try:
                conn = self._conn_manager.get_connection(
                    endpoint=self._address.hostname, auth=mgmt_auth
                )
                mgmt_client.open(connection=conn)
                while not mgmt_client.client_ready():
                    time.sleep(0.05)

                cast(Dict[Union[str, bytes], Any], mgmt_msg.application_properties)["security_token"] = (
                    self._amqp_transport.get_updated_token(mgmt_auth)
                )
                status_code, description, response = self._amqp_transport.mgmt_client_request(
                    mgmt_client,
                    mgmt_msg,
                    operation=READ_OPERATION,
                    operation_type=op_type,
                    status_code_field=MGMT_STATUS_CODE,
                    description_fields=MGMT_STATUS_DESC,
                )
                status_code = int(status_code)
                if description and isinstance(description, bytes):
                    description = description.decode("utf-8")
                if status_code < 400:
                    return response
                raise self._amqp_transport.get_error(status_code, description)
            except Exception as exception:  # pylint: disable=broad-except
                # If optional dependency is not installed, do not retry.
                if isinstance(exception, ImportError):
                    raise exception
                # is_consumer=True passed in here, ALTHOUGH this method is shared by the producer and consumer.
                # is_consumer will only be checked if FileNotFoundError is raised by self.mgmt_client.open() due to
                # invalid/non-existent connection_verify filepath. The producer will encounter the FileNotFoundError
                # when opening the SendClient, so is_consumer=True will not be passed to amqp_transport.handle_exception
                # there. This is for uamqp exception parity, which raises FileNotFoundError in the consumer and
                # EventHubError in the producer. TODO: Remove `is_consumer` kwarg when resolving issue #27128.
                last_exception = self._amqp_transport._handle_exception(  # pylint: disable=protected-access
                    exception, self, is_consumer=True
                )
                self._backoff(retried_times=retried_times, last_exception=last_exception)
                retried_times += 1
                if retried_times > self._config.max_retries:
                    _LOGGER.info("%r returns an exception %r", self._container_id, last_exception)
                    raise last_exception from None
            finally:
                mgmt_client.close()

    def _get_eventhub_properties(self) -> Dict[str, Any]:
        mgmt_msg = self._amqp_transport.build_message(application_properties={"name": self.eventhub_name})
        response = self._management_request(mgmt_msg, op_type=MGMT_OPERATION)
        output = {}
        eh_info: Dict[bytes, Any] = response.value
        if eh_info:
            output["eventhub_name"] = eh_info[b"name"].decode("utf-8")
            output["created_at"] = utc_from_timestamp(float(eh_info[b"created_at"]) / 1000)
            output["partition_ids"] = [p.decode("utf-8") for p in eh_info[b"partition_ids"]]
        return output

    def _get_partition_ids(self) -> List[str]:
        return self._get_eventhub_properties()["partition_ids"]

    def _get_partition_properties(self, partition_id: str) -> Dict[str, Any]:
        mgmt_msg = self._amqp_transport.build_message(
            application_properties={
                "name": self.eventhub_name,
                "partition": partition_id,
            }
        )
        response = self._management_request(mgmt_msg, op_type=MGMT_PARTITION_OPERATION)
        partition_info: Dict[bytes, Union[bytes, int]] = response.value
        output: Dict[str, Any] = {}
        if partition_info:
            output["eventhub_name"] = cast(bytes, partition_info[b"name"]).decode("utf-8")
            output["id"] = cast(bytes, partition_info[b"partition"]).decode("utf-8")
            output["beginning_sequence_number"] = cast(int, partition_info[b"begin_sequence_number"])
            output["last_enqueued_sequence_number"] = cast(int, partition_info[b"last_enqueued_sequence_number"])
            output["last_enqueued_offset"] = cast(bytes, partition_info[b"last_enqueued_offset"]).decode("utf-8")
            output["is_empty"] = partition_info[b"is_partition_empty"]
            output["last_enqueued_time_utc"] = utc_from_timestamp(
                float(cast(int, partition_info[b"last_enqueued_time_utc"]) / 1000)
            )
        return output

    def _close(self) -> None:
        self._conn_manager.close_connection()


class ConsumerProducerMixin:

    def __init__(self) -> None:
        self._handler: Union[
            uamqp_AMQPRecieveClient, pyamqp_AMQPRecieveClient, uamqp_AMQPSendClient, pyamqp_AMQPSendClient
        ]
        self._client: Union[EventHubConsumerClient, EventHubProducerClient]
        self._amqp_transport: "AmqpTransport"
        self._max_message_size_on_link: Optional[int] = None

    def __enter__(self) -> ConsumerProducerMixin:
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.close()

    def _create_handler(self, auth):
        pass

    def _check_closed(self):
        if self.closed:
            raise ClientClosedError(f"{self._name} has been closed. Please create a new one to handle event data.")

    def _open(self) -> bool:
        """Open the EventHubConsumer/EventHubProducer using the supplied connection.

        :return: Whether the EventHubConsumer/EventHubProducer is ready to use.
        :rtype: bool
        """
        # pylint: disable=protected-access
        if not self.running:
            if self._handler:
                self._handler.close()
            auth = self._client._create_auth(auth_uri=self._client._auth_uri)
            self._create_handler(auth)
            conn = self._client._conn_manager.get_connection(  # pylint: disable=protected-access
                endpoint=self._client._address.hostname, auth=auth
            )
            self._handler.open(connection=conn)
            while not self._handler.client_ready():
                time.sleep(0.05)
            self._max_message_size_on_link = (
                self._amqp_transport.get_remote_max_message_size(self._handler)
                or self._amqp_transport.MAX_MESSAGE_LENGTH_BYTES
            )
            self.running: bool = True
            return True
        return False

    def _close_handler(self) -> None:
        if self._handler:
            self._handler.close()  # close the link (sharing connection) or connection (not sharing)
        self.running = False

    def _close_connection(self):
        self._close_handler()
        self._client._conn_manager.reset_connection_if_broken()  # pylint: disable=protected-access

    def _handle_exception(self, exception, *, is_consumer=False):
        exception = self._amqp_transport.check_timeout_exception(self, exception)
        return self._amqp_transport._handle_exception(  # pylint: disable=protected-access
            exception, self, is_consumer=is_consumer
        )

    def _do_retryable_operation(  # pylint:disable=inconsistent-return-statements
        self, operation, timeout=None, **kwargs
    ):
        # pylint:disable=protected-access
        timeout_time = (time.time() + timeout) if timeout else None
        retried_times = 0
        last_exception = kwargs.pop("last_exception", None)
        operation_need_param = kwargs.pop("operation_need_param", True)
        max_retries = self._client._config.max_retries  # pylint:disable=protected-access

        while retried_times <= max_retries:
            try:
                if operation_need_param:
                    return operation(
                        timeout_time=timeout_time,
                        last_exception=last_exception,
                        **kwargs,
                    )
                return operation()
            except Exception as exception:  # pylint:disable=broad-except
                # If optional dependency is not installed, do not retry.
                if isinstance(exception, ImportError):
                    raise exception
                last_exception = self._handle_exception(exception)
                self._client._backoff(
                    retried_times=retried_times,
                    last_exception=last_exception,
                    timeout_time=timeout_time,
                    entity_name=self._name,
                )
                retried_times += 1
                if retried_times > max_retries:
                    _LOGGER.info(
                        "%r operation has exhausted retry. Last exception: %r.",
                        self._name,
                        last_exception,
                    )
                    raise last_exception from None

    def close(self) -> None:
        """
        Close down the handler. If the handler has already closed,
        this will be a no op.
        """
        self._close_handler()
        self.closed = True



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------

import time
from functools import partial

from typing import Any, Callable, NamedTuple, Optional, Tuple, Union
from .sasl import SASLAnonymousCredential, SASLPlainCredential
from .utils import generate_sas_token

from .constants import (
    AUTH_DEFAULT_EXPIRATION_SECONDS,
    TOKEN_TYPE_JWT,
    TOKEN_TYPE_SASTOKEN,
    AUTH_TYPE_CBS,
    AUTH_TYPE_SASL_PLAIN,
)


class AccessToken(NamedTuple):
    token: Union[str, bytes]
    expires_on: int


def _generate_sas_access_token(
    auth_uri: str, sas_name: str, sas_key: str, expiry_in: float = AUTH_DEFAULT_EXPIRATION_SECONDS
) -> AccessToken:
    expires_on = int(time.time() + expiry_in)
    token = generate_sas_token(auth_uri, sas_name, sas_key, expires_on)
    return AccessToken(token, expires_on)


class SASLPlainAuth:
    # TODO:
    #  1. naming decision, suffix with Auth vs Credential
    auth_type = AUTH_TYPE_SASL_PLAIN

    def __init__(self, authcid: str, passwd: str, authzid: Optional[str] = None):
        self.sasl = SASLPlainCredential(authcid, passwd, authzid)


class _CBSAuth:
    # TODO:
    #  1. naming decision, suffix with Auth vs Credential
    auth_type = AUTH_TYPE_CBS

    def __init__(  # pylint: disable=unused-argument
        self,
        uri: str,
        audience: str,
        token_type: Union[str, bytes],
        get_token: Callable[[], AccessToken],
        *,
        expires_in: Optional[float] = AUTH_DEFAULT_EXPIRATION_SECONDS,
        expires_on: Optional[float] = None,
        **kwargs: Any
    ):
        """
        CBS authentication using JWT tokens.

        :param uri: The AMQP endpoint URI. This must be provided as
         a decoded string.
        :type uri: str
        :param audience: The token audience field. For SAS tokens
         this is usually the URI.
        :type audience: str
        :param get_token: The callback function used for getting and refreshing
         tokens. It should return a valid jwt token each time it is called.
        :type get_token: callable object
        :param token_type: The type field of the token request.
         Default value is `"jwt"`.
        :type token_type: str

        """
        self.sasl = SASLAnonymousCredential()
        self.uri = uri
        self.audience = audience
        self.token_type = token_type
        self.get_token = get_token
        self.expires_in = expires_in
        self.expires_on = expires_on

    @staticmethod
    def _set_expiry(expires_in: Optional[float] = None, expires_on: Optional[float] = None) -> Tuple[float, float]:
        if not expires_on and not expires_in:
            raise ValueError("Must specify either 'expires_on' or 'expires_in'.")

        expires_in_interval: float = 0 if expires_in is None else expires_in
        expires_on_time: float = 0 if expires_on is None else expires_on

        if not expires_on and expires_in:
            expires_on_time = time.time() + expires_in
        elif expires_on and not expires_in:
            expires_in_interval = expires_on - time.time()
            if expires_in_interval < 1:
                raise ValueError("Token has already expired.")

        return expires_in_interval, expires_on_time


class JWTTokenAuth(_CBSAuth):
    # TODO:
    #  1. naming decision, suffix with Auth vs Credential
    def __init__(
        self,
        uri: str,
        audience: str,
        get_token: Callable[..., AccessToken],
        *,
        token_type: Union[str, bytes] = TOKEN_TYPE_JWT,
        **kwargs: Any
    ):
        """
        CBS authentication using JWT tokens.

        :param uri: The AMQP endpoint URI. This must be provided as
         a decoded string.
        :type uri: str
        :param audience: The token audience field. For SAS tokens
         this is usually the URI.
        :type audience: str
        :param get_token: The callback function used for getting and refreshing
         tokens. It should return a valid jwt token each time it is called.
        :type get_token: callable object
        :param token_type: The type field of the token request.
         Default value is `"jwt"`.
        :type token_type: str

        """
        super(JWTTokenAuth, self).__init__(uri, audience, token_type, get_token)
        self.get_token = get_token


class SASTokenAuth(_CBSAuth):
    # TODO:
    #  1. naming decision, suffix with Auth vs Credential
    def __init__(
        self,
        uri: str,
        audience: str,
        username: str,
        password: str,
        *,
        expires_in: Optional[float] = AUTH_DEFAULT_EXPIRATION_SECONDS,
        expires_on: Optional[float] = None,
        token_type: Union[str, bytes] = TOKEN_TYPE_SASTOKEN,
        **kwargs: Any
    ):
        """
        CBS authentication using SAS tokens.

        :param uri: The AMQP endpoint URI. This must be provided as
         a decoded string.
        :type uri: str
        :param audience: The token audience field. For SAS tokens
         this is usually the URI.
        :type audience: str
        :param username: The SAS token username, also referred to as the key
         name or policy name. This can optionally be encoded into the URI.
        :type username: str
        :param password: The SAS token password, also referred to as the key.
         This can optionally be encoded into the URI.
        :type password: str
        :param expires_in: The total remaining seconds until the token
         expires.
        :type expires_in: int
        :param expires_on: The timestamp at which the SAS token will expire
         formatted as seconds since epoch.
        :type expires_on: float
        :param token_type: The type field of the token request.
         Default value is `"servicebus.windows.net:sastoken"`.
        :type token_type: str

        """
        self.username = username
        self.password = password
        expires_in, expires_on = self._set_expiry(expires_in, expires_on)
        self.get_token = partial(_generate_sas_access_token, uri, username, password, expires_in)
        super(SASTokenAuth, self).__init__(
            uri, audience, token_type, self.get_token, expires_in=expires_in, expires_on=expires_on
        )



# coding=utf-8
# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
import pytest
from threading import Lock
from azure.eventhub._buffered_producer._partition_resolver import generate_hash_code, PartitionResolver


class TestPartitionResolver:
    def test_partition_key(self):
        input = {
            "7": -15263,
            "7149583486996073602": 12977,
            "FWfAT": -22341,
            "sOdeEAsyQoEuEFPGerWO": -6503,
            "FAyAIctPeCgmiwLKbJcyswoHglHVjQdvtBowLACDNORsYvOcLddNJYDmhAVkbyLOrHTKLneMNcbgWVlasVywOByANjs": 5226,
            "1XYM6!(7(lF5wq4k4m*e$Nc!1ezLJv*1YK1Y-C^*&B$O)lq^iUkG(TNzXG;Zi#z2Og*Qq0#^*k):vXh$3,C7We7%W0meJ;b3,rQCg^J;^twXgs5E$$hWKxqp": 23950,
            "E(x;RRIaQcJs*P;D&jTPau-4K04oqr:lF6Z):ERpo&;9040qyV@G1_c9mgOs-8_8/10Fwa-7b7-yP!T-!IH&968)FWuI;(^g$2fN;)HJ^^yTn:": -29304,
            "!c*_!I@1^c": 15372,
            "p4*!jioeO/z-!-;w:dh": -3104,
            "$0cb": 26269,
            "-4189260826195535198": 453,
        }

        for k, v in input.items():
            assert generate_hash_code(k) == v

    @pytest.mark.parametrize("partition_cnt", [1, 2, 16, 32, 256])
    def test_basic_round_robin(self, partition_cnt):
        partitions = [str(i) for i in range(partition_cnt)]
        pr = PartitionResolver(partitions)
        for i in range(2 * partition_cnt):
            expected = str(i % partition_cnt)
            real = pr.get_next_partition_id()
            assert expected == real

    @pytest.mark.parametrize("partition_cnt", [1, 2, 16, 32, 256])
    def test_concurrent_round_robin_fairly(self, partition_cnt):
        partitions = [str(i) for i in range(partition_cnt)]
        pr = PartitionResolver(partitions)
        exc = ThreadPoolExecutor()

        dic = defaultdict(int)
        lock = Lock()

        def gen_pid():
            pid = pr.get_next_partition_id()
            with lock:
                dic[pid] += 1

        for i in range(5 * partition_cnt):
            exc.submit(gen_pid)

        exc.shutdown()
        assert len(dic) == partition_cnt
        for i in range(partition_cnt):
            assert dic[str(i)] == 5



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from __future__ import annotations
from typing import TYPE_CHECKING, Any, Dict, Optional, Union
from asyncio import Lock # pylint:disable=do-not-import-asyncio

from .._connection_manager import _ConnectionMode
from .._constants import TransportType

if TYPE_CHECKING:
    from .._pyamqp.aio._authentication_async import JWTTokenAuthAsync
    from .._pyamqp.aio._connection_async import Connection as ConnectionAsync
    from ._transport._base_async import AmqpTransportAsync

    try:
        from uamqp.authentication import JWTTokenAsync as uamqp_JWTTokenAuthAsync
        from uamqp.async_ops import ConnectionAsync as uamqp_ConnectionAsync
    except ImportError:
        uamqp_JWTTokenAuthAsync = None
        uamqp_ConnectionAsync = None

    try:
        from typing_extensions import Protocol
    except ImportError:
        Protocol = object  # type: ignore

    class ConnectionManager(Protocol):
        async def get_connection(
            self,
            *,
            endpoint: str,
            auth: Optional[Union[uamqp_JWTTokenAuthAsync, JWTTokenAuthAsync]] = None,
        ) -> Union[ConnectionAsync, uamqp_ConnectionAsync]:
            pass

        async def close_connection(self) -> None:
            pass

        async def reset_connection_if_broken(self) -> None:
            pass


class _SharedConnectionManager:  # pylint:disable=too-many-instance-attributes
    def __init__(
        self,
        *,
        container_id: Optional[str] = None,
        custom_endpoint_address: Optional[str] = None,
        debug: bool = False,
        error_policy: Optional[Any] = None,
        properties: Optional[Dict[str, Any]] = None,
        encoding: str = "UTF-8",
        transport_type: TransportType = TransportType.Amqp,
        http_proxy: Optional[str] = None,
        max_frame_size: int,
        channel_max: int,
        idle_timeout: Optional[float],
        remote_idle_timeout_empty_frame_send_ratio: float,
        amqp_transport: AmqpTransportAsync,
        **kwargs: Any,
    ) -> None:
        self._loop = kwargs.get("loop")
        self._lock = Lock()
        self._conn: Optional[Union[uamqp_ConnectionAsync, ConnectionAsync]] = None

        self._container_id = container_id
        self._custom_endpoint_address = custom_endpoint_address
        self._debug = debug
        self._error_policy = error_policy
        self._properties = properties
        self._encoding = encoding
        self._transport_type = transport_type
        self._http_proxy = http_proxy
        self._max_frame_size = max_frame_size
        self._channel_max = channel_max
        self._idle_timeout = idle_timeout
        self._remote_idle_timeout_empty_frame_send_ratio = remote_idle_timeout_empty_frame_send_ratio
        self._amqp_transport: AmqpTransportAsync = amqp_transport

    async def get_connection(
        self,
        *,
        endpoint: str,
        auth: Optional[Union[JWTTokenAuthAsync, uamqp_JWTTokenAuthAsync]] = None,
    ) -> Union[ConnectionAsync, uamqp_ConnectionAsync]:
        async with self._lock:
            if self._conn is None:
                self._conn = self._amqp_transport.create_connection_async(
                    endpoint=endpoint,
                    auth=auth,
                    custom_endpoint_address=self._custom_endpoint_address,
                    container_id=self._container_id,
                    max_frame_size=self._max_frame_size,
                    channel_max=self._channel_max,
                    idle_timeout=self._idle_timeout,
                    properties=self._properties,
                    remote_idle_timeout_empty_frame_send_ratio=self._remote_idle_timeout_empty_frame_send_ratio,
                    error_policy=self._error_policy,
                    debug=self._debug,
                    loop=self._loop,
                    encoding=self._encoding,
                )
            return self._conn

    async def close_connection(self) -> None:
        async with self._lock:
            if self._conn:
                await self._amqp_transport.close_connection_async(self._conn)
            self._conn = None

    async def reset_connection_if_broken(self) -> None:
        async with self._lock:
            conn_state = self._amqp_transport.get_connection_state(self._conn)
            if self._conn and conn_state in self._amqp_transport.CONNECTION_CLOSING_STATES:
                self._conn = None


class _SeparateConnectionManager:
    def __init__(self, **kwargs) -> None:
        pass

    async def get_connection(
        self,
        *,
        endpoint: str,
        auth: Optional[Union[JWTTokenAuthAsync, uamqp_JWTTokenAuthAsync]] = None,
    ) -> None:
        pass  # return None

    async def close_connection(self) -> None:
        pass

    async def reset_connection_if_broken(self) -> None:
        pass


def get_connection_manager(**kwargs) -> "ConnectionManager":
    connection_mode = kwargs.get("connection_mode", _ConnectionMode.SeparateConnection)  # type: ignore
    if connection_mode == _ConnectionMode.ShareConnection:
        return _SharedConnectionManager(**kwargs)
    return _SeparateConnectionManager(**kwargs)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import uuid
import logging
from typing import Optional, Union

from .._decode import decode_payload
from ._link_async import Link
from ..constants import LinkState, Role
from ..performatives import (
    TransferFrame,
    DispositionFrame,
)
from ..outcomes import Received, Accepted, Rejected, Released, Modified
from ..error import AMQPException, ErrorCondition


_LOGGER = logging.getLogger(__name__)


class ReceiverLink(Link):
    def __init__(self, session, handle, source_address, **kwargs):
        name = kwargs.pop("name", None) or str(uuid.uuid4())
        role = Role.Receiver
        if "target_address" not in kwargs:
            kwargs["target_address"] = "receiver-link-{}".format(name)
        super(ReceiverLink, self).__init__(session, handle, name, role, source_address=source_address, **kwargs)
        self._on_transfer = kwargs.pop("on_transfer")
        self._received_payload = bytearray()
        self._first_frame = None
        self._received_delivery_tags = set()

    @classmethod
    def from_incoming_frame(cls, session, handle, frame):
        # TODO: Assuming we establish all links for now...
        # check link_create_from_endpoint in C lib
        raise NotImplementedError("Pending")

    async def _process_incoming_message(self, frame, message):
        try:
            return await self._on_transfer(frame, message)
        except Exception as e:  # pylint: disable=broad-except
            _LOGGER.error("Transfer callback function failed with error: %r", e, extra=self.network_trace_params)
        return None

    async def _incoming_attach(self, frame):
        await super(ReceiverLink, self)._incoming_attach(frame)
        if frame[9] is None:  # initial_delivery_count
            _LOGGER.info("Cannot get initial-delivery-count. Detaching link", extra=self.network_trace_params)
            await self._set_state(LinkState.DETACHED)  # TODO: Send detach now?
        self.delivery_count = frame[9]
        self.current_link_credit = self.link_credit
        await self._outgoing_flow()

    async def _incoming_transfer(self, frame):
        if self.network_trace:
            _LOGGER.debug("<- %r", TransferFrame(payload=b"***", *frame[:-1]), extra=self.network_trace_params)
        self.received_delivery_id = frame[1]  # delivery_id
        # If more is false --> this is the last frame of the message
        if not frame[5]:
            self.delivery_count += 1
            self.current_link_credit -= 1
        if self.received_delivery_id is not None:
            self._first_frame = frame
        if not self.received_delivery_id and not self._received_payload:
            pass  # TODO: delivery error
        if self._received_payload or frame[5]:  # more
            self._received_payload.extend(frame[11])
        if not frame[5]:
            self._received_delivery_tags.add(self._first_frame[2])
            if self._received_payload:
                message = decode_payload(memoryview(self._received_payload))
                self._received_payload = bytearray()
            else:
                message = decode_payload(frame[11])
            delivery_state = await self._process_incoming_message(self._first_frame, message)
            if not frame[4] and delivery_state:  # settled
                await self._outgoing_disposition(
                    first=self._first_frame[1],
                    last=self._first_frame[1],
                    delivery_tag=self._first_frame[2],
                    settled=True,
                    state=delivery_state,
                    batchable=None,
                )

    async def _wait_for_response(self, wait: Union[bool, float]) -> None:
        if wait is True:
            await self._session._connection.listen(wait=False)  # pylint: disable=protected-access
            if self.state == LinkState.ERROR:
                if self._error:
                    raise self._error
        elif wait:
            await self._session._connection.listen(wait=wait)  # pylint: disable=protected-access
            if self.state == LinkState.ERROR:
                if self._error:
                    raise self._error

    async def _outgoing_disposition(
        self,
        first: int,
        last: Optional[int],
        delivery_tag: bytes,
        settled: Optional[bool],
        state: Optional[Union[Received, Accepted, Rejected, Released, Modified]],
        batchable: Optional[bool],
    ):
        disposition_frame = DispositionFrame(
            role=self.role, first=first, last=last, settled=settled, state=state, batchable=batchable
        )
        if delivery_tag not in self._received_delivery_tags:
            raise AMQPException(condition=ErrorCondition.IllegalState, description="Delivery tag not found.")

        if self.network_trace:
            _LOGGER.debug("-> %r", DispositionFrame(*disposition_frame), extra=self.network_trace_params)
        await self._session._outgoing_disposition(disposition_frame)  # pylint: disable=protected-access
        self._received_delivery_tags.remove(delivery_tag)

    async def attach(self):
        await super().attach()
        self._received_payload = bytearray()

    async def send_disposition(
        self,
        *,
        wait: Union[bool, float] = False,
        first_delivery_id: int,
        last_delivery_id: Optional[int] = None,
        delivery_tag: bytes,
        settled: Optional[bool] = None,
        delivery_state: Optional[Union[Received, Accepted, Rejected, Released, Modified]] = None,
        batchable: Optional[bool] = None
    ):
        if self._is_closed:
            raise ValueError("Link already closed.")
        await self._outgoing_disposition(
            first_delivery_id, last_delivery_id, delivery_tag, settled, delivery_state, batchable
        )
        if not settled:
            await self._wait_for_response(wait)



import sys
import pytest
import asyncio
from unittest.mock import patch
from azure.eventhub._pyamqp.aio._transport_async import WebSocketTransportAsync


# class WebsocketException(unittest.TestCase):
@pytest.mark.asyncio
async def test_websocket_aiohttp_exception_async():
    import aiohttp
    with patch.object(aiohttp.ClientSession, "ws_connect", side_effect=aiohttp.ClientOSError):
        transport = WebSocketTransportAsync(host="my_host")
        with pytest.raises(aiohttp.ClientOSError):
            await transport.connect()



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Example to show sending, receiving and parsing amqp annotated message(s) to Event Hubs.
"""

import os
import asyncio
from azure.eventhub import TransportType
from azure.eventhub.aio import EventHubProducerClient, EventHubConsumerClient
from azure.eventhub.amqp import AmqpAnnotatedMessage, AmqpMessageBodyType
from azure.identity.aio import DefaultAzureCredential

FULLY_QUALIFIED_NAMESPACE = os.environ["EVENT_HUB_HOSTNAME"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]


async def send_data_message(producer):
    data_body = [b"aa", b"bb", b"cc"]
    application_properties = {"body_type": "data"}
    delivery_annotations = {"delivery_annotation_key": "value"}
    data_message = AmqpAnnotatedMessage(
        data_body=data_body, delivery_annotations=delivery_annotations, application_properties=application_properties
    )
    batch = await producer.create_batch()
    batch.add(data_message)
    await producer.send_batch(batch)
    print("Message of data body sent.")


async def send_sequence_message(producer):
    sequence_body = [b"message", 123.456, True]
    footer = {"footer_key": "footer_value"}
    properties = {"subject": "sequence"}
    application_properties = {"body_type": "sequence"}
    sequence_message = AmqpAnnotatedMessage(
        sequence_body=sequence_body, footer=footer, properties=properties, application_properties=application_properties
    )
    await producer.send_batch([sequence_message])
    print("Message of sequence body sent.")


async def send_value_message(producer):
    value_body = {b"key": [-123, b"data", False]}
    header = {"priority": 10}
    annotations = {"annotation_key": "value"}
    application_properties = {"body_type": "value"}
    value_message = AmqpAnnotatedMessage(
        value_body=value_body, header=header, annotations=annotations, application_properties=application_properties
    )
    await producer.send_batch([value_message])
    print("Message of value body sent.")


async def on_event(partition_context, event):
    # Put your code here.
    # If the operation is i/o intensive, multi-thread will have better performance.
    print("Received event from partition: {}".format(partition_context.partition_id))
    raw_amqp_message = event.raw_amqp_message
    if raw_amqp_message.body_type == AmqpMessageBodyType.DATA:
        print("Message of data body received. Body is:")
        for data_section in raw_amqp_message.body:
            print(data_section)
    elif raw_amqp_message.body_type == AmqpMessageBodyType.SEQUENCE:
        print("Message of sequence body received. Body is:")
        for sequence_section in raw_amqp_message.body:
            print(sequence_section)
    elif raw_amqp_message.body_type == AmqpMessageBodyType.VALUE:
        print("Message of value body received. Body is:")
        print(raw_amqp_message.body)


async def receive_and_parse_message(consumer):
    async with consumer:
        try:
            await consumer.receive(
                on_event=on_event,
                starting_position="-1",  # "-1" is from the beginning of the partition.
            )
        except KeyboardInterrupt:
            print("Stopped receiving.")


async def main():
    # Send AmqpAnnotatedMessage
    producer = EventHubProducerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        eventhub_name=EVENTHUB_NAME,
        credential=DefaultAzureCredential(),
        transport_type=TransportType.AmqpOverWebsocket,
    )
    async with producer:
        await send_data_message(producer)
        await send_sequence_message(producer)
        await send_value_message(producer)

    # Receive
    consumer = EventHubConsumerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        credential=DefaultAzureCredential(),
        consumer_group="$Default",
        eventhub_name=EVENTHUB_NAME,
        transport_type=TransportType.AmqpOverWebsocket,
    )
    await receive_and_parse_message(consumer)


if __name__ == "__main__":
    asyncio.run(main())



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
An example to show sending and receiving events behind a proxy.
"""
import os
from typing import Dict, Union
from azure.eventhub import EventData, EventHubConsumerClient, EventHubProducerClient
from azure.identity import DefaultAzureCredential

FULLY_QUALIFIED_NAMESPACE = os.environ["EVENT_HUB_HOSTNAME"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]

HTTP_PROXY: Dict[str, Union[str, int]] = {
    "proxy_hostname": "127.0.0.1",  # proxy hostname.
    "proxy_port": 3128,  # proxy port.
    "username": "admin",  # username used for proxy authentication if needed.
    "password": "123456",  # password used for proxy authentication if needed.
}


def on_event(partition_context, event):
    # Put your code here.
    print("received event from partition: {}.".format(partition_context.partition_id))
    print(event)


consumer_client = EventHubConsumerClient(
    fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
    consumer_group="$Default",
    eventhub_name=EVENTHUB_NAME,
    credential=DefaultAzureCredential(),
    http_proxy=HTTP_PROXY,
)
producer_client = EventHubProducerClient(
    fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
    eventhub_name=EVENTHUB_NAME,
    credential=DefaultAzureCredential(),
    http_proxy=HTTP_PROXY,
)

with producer_client:
    event_data_batch = producer_client.create_batch(max_size_in_bytes=10000)
    while True:
        try:
            event_data_batch.add(EventData("Message inside EventBatchData"))
        except ValueError:
            # EventDataBatch object reaches max_size.
            # New EventDataBatch object can be created here to send more data.
            break
    producer_client.send_batch(event_data_batch)
    print("Finished sending.")

with consumer_client:
    consumer_client.receive(on_event=on_event)
    print("Finished receiving.")



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Example to demonstrate utilizing AzureNamedKeyCredential to authenticate with Event Hubs asynchronously.
"""

import os
import asyncio
import time
from azure.core.credentials import AzureNamedKeyCredential
from azure.eventhub.aio import EventHubProducerClient
from azure.eventhub import EventData

# Target namespace and hub must also be specified.
FULLY_QUALIFIED_NAMESPACE = os.environ["EVENT_HUB_HOSTNAME"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]

EVENTHUB_POLICY_NAME = os.environ["EVENT_HUB_SAS_POLICY"]
EVENTHUB_KEY = os.environ["EVENT_HUB_SAS_KEY"]

credential = AzureNamedKeyCredential(EVENTHUB_POLICY_NAME, EVENTHUB_KEY)

producer_client = EventHubProducerClient(
    fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
    eventhub_name=EVENTHUB_NAME,
    credential=credential,
    logging_enable=True,
)

start_time = time.time()


async def authenticate_with_named_key():
    async with producer_client:
        event_data_batch = await producer_client.create_batch()
        event_data_batch.add(EventData("Single message"))
        await producer_client.send_batch(event_data_batch)


start_time = time.time()
asyncio.run(authenticate_with_named_key())
print("Send messages in {} seconds.".format(time.time() - start_time))



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from typing import Any, Dict, Optional
from ._mixin import DictMixin
from ._client_base import _parse_conn_str


class EventHubConnectionStringProperties(DictMixin):
    """
    Properties of a connection string.
    """

    def __init__(  # pylint: disable=unused-argument
        self,
        *,
        fully_qualified_namespace: str,
        endpoint: str,
        eventhub_name: Optional[str] = None,
        shared_access_signature: Optional[str] = None,
        shared_access_key_name: Optional[str] = None,
        shared_access_key: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        self._fully_qualified_namespace: str = fully_qualified_namespace
        self._endpoint: str = endpoint
        self._eventhub_name: Optional[str] = eventhub_name
        self._shared_access_signature: Optional[str] = shared_access_signature
        self._shared_access_key_name: Optional[str] = shared_access_key_name
        self._shared_access_key: Optional[str] = shared_access_key

    @property
    def fully_qualified_namespace(self) -> str:
        """The fully qualified host name for the Event Hubs namespace.
        The namespace format is: `<yournamespace>.servicebus.windows.net`.

        :rtype: str
        """
        return self._fully_qualified_namespace

    @property
    def endpoint(self) -> str:
        """The endpoint for the Event Hubs resource. In the format sb://<FQDN>/

        :rtype: str
        """
        return self._endpoint

    @property
    def eventhub_name(self) -> Optional[str]:
        """Optional. The name of the Event Hub, represented by `EntityPath` in the connection string.

        :rtype: str
        """
        return self._eventhub_name

    @property
    def shared_access_signature(self) -> Optional[str]:
        """
        This can be provided instead of the shared_access_key_name and the shared_access_key.

        :rtype: str
        """
        return self._shared_access_signature

    @property
    def shared_access_key_name(self) -> Optional[str]:
        """
        The name of the shared_access_key. This must be used along with the shared_access_key.

        :rtype: str
        """
        return self._shared_access_key_name

    @property
    def shared_access_key(self) -> Optional[str]:
        """
        The shared_access_key can be used along with the shared_access_key_name as a credential.

        :rtype: str
        """
        return self._shared_access_key


def parse_connection_string(conn_str: str) -> "EventHubConnectionStringProperties":
    """Parse the connection string into a properties bag containing its component parts.

    :param conn_str: The connection string that has to be parsed.
    :type conn_str: str
    :return: A properties bag containing the parsed connection string.
    :rtype: ~azure.eventhub.EventHubConnectionStringProperties
    """
    fully_qualified_namespace, policy, key, entity, signature, emulator = _parse_conn_str(conn_str, check_case=True)[
        :-1
    ]
    endpoint = "sb://" + fully_qualified_namespace + "/"
    props: Dict[str, Any] = {
        "fully_qualified_namespace": fully_qualified_namespace,
        "endpoint": endpoint,
        "eventhub_name": entity,
        "shared_access_signature": signature,
        "shared_access_key_name": policy,
        "shared_access_key": key,
        "emulator": emulator,
    }
    return EventHubConnectionStringProperties(**props)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from azure.core.utils import parse_connection_string
from uamqp.client import ReceiveClient
from uamqp.async_ops.client_async import ReceiveClientAsync
from uamqp import authentication

from devtools_testutils.perfstress_tests import EventPerfTest


class UamqpReceiveEventTest(EventPerfTest):
    def __init__(self, arguments):
        super().__init__(arguments)
        live_eventhub_config = self._get_eh_connection_config()
        uri = "{}{}".format(live_eventhub_config["hostname"], live_eventhub_config["event_hub"])
        source = "{}{}/ConsumerGroups/{}/Partitions/{}".format(
            live_eventhub_config["hostname"],
            live_eventhub_config["event_hub"],
            live_eventhub_config["consumer_group"],
            live_eventhub_config["partition"],
        ).replace("sb://", "amqps://")

        # Setup service clients
        self.receive_client = ReceiveClient(
            source,
            auth=authentication.SASTokenAuth.from_shared_access_key(
                uri, live_eventhub_config["key_name"], live_eventhub_config["access_key"]
            ),
            timeout=0,
            debug=False,
        )
        self.async_receive_client = ReceiveClientAsync(
            source,
            auth=authentication.SASTokenAsync.from_shared_access_key(
                uri, live_eventhub_config["key_name"], live_eventhub_config["access_key"]
            ),
            timeout=0,
            debug=False,
        )

    def _get_eh_connection_config(self):
        connection_string = self.get_from_env("AZURE_EVENTHUB_CONNECTION_STRING")
        auth_info = parse_connection_string(connection_string)
        config = {}
        config["hostname"] = auth_info["endpoint"]
        config["event_hub"] = self.get_from_env("AZURE_EVENTHUB_NAME")
        config["key_name"] = auth_info["sharedaccesskeyname"]
        config["access_key"] = auth_info["sharedaccesskey"]
        config["consumer_group"] = "$Default"
        config["partition"] = "0"
        return config

    def process_event_sync(self, message):
        try:
            message.annotations
            message.properties
            message.get_data()
            message.header
            message.delivery_annotations
            self.event_raised_sync()
        except Exception as e:
            self.error_raised_sync(e)

    def start_events_sync(self) -> None:
        """
        Start the process for receiving events.
        """
        self.receive_client.receive_messages(self.process_event_sync)

    def stop_events_sync(self) -> None:
        """
        Stop the process for receiving events.
        """
        self.receive_client.close()

    async def start_events_async(self) -> None:
        """
        Start the process for receiving events.
        """
        await self.async_receive_client.receive_messages_async(self.process_event_sync)

    async def stop_events_async(self) -> None:
        """
        Stop the process for receiving events.
        """
        await self.async_receive_client.close_async()



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Examples to show sending events in buffered mode to an Event Hub asynchronously.
"""

import time
import asyncio
import os

from azure.eventhub.aio import EventHubProducerClient
from azure.eventhub import EventData
from azure.identity.aio import DefaultAzureCredential

FULLY_QUALIFIED_NAMESPACE = os.environ["EVENT_HUB_HOSTNAME"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]


async def on_success(events, pid):
    # sending succeeded
    print(events, pid)


async def on_error(events, pid, error):
    # sending failed
    print(events, pid, error)


async def run():

    producer = EventHubProducerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        eventhub_name=EVENTHUB_NAME,
        credential=DefaultAzureCredential(),
        buffered_mode=True,
        on_success=on_success,
        on_error=on_error,
    )

    # exiting the context manager will automatically call flush
    async with producer:
        # single events will be batched automatically
        for i in range(10):
            # the method returning indicates the event has been enqueued to the buffer
            await producer.send_event(EventData("Single data {}".format(i)))

        batch = await producer.create_batch()
        for i in range(10):
            batch.add(EventData("Single data in batch {}".format(i)))
        # alternatively, you can enqueue an EventDataBatch object to the buffer
        await producer.send_batch(batch)

        # calling flush sends out the events in the buffer immediately
        await producer.flush()


start_time = time.time()
asyncio.run(run())
print("Send messages in {} seconds.".format(time.time() - start_time))



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Examples to show how to create EventHubProducerClient and EventHubConsumerClient that connect to custom endpoint.
"""

import os
from azure.eventhub import EventHubProducerClient, EventHubConsumerClient, EventData
from azure.identity import DefaultAzureCredential

FULLY_QUALIFIED_NAMESPACE = os.environ["EVENT_HUB_HOSTNAME"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]
# The custom endpoint address to use for establishing a connection to the Event Hubs service,
# allowing network requests to be routed through any application gateways
# or other paths needed for the host environment.
CUSTOM_ENDPOINT_ADDRESS = "sb://<custom_endpoint_hostname>:<custom_endpoint_port>"
# The optional absolute path to the custom certificate file used by client to authenticate the
# identity of the connection endpoint in the case that endpoint has its own issued CA.
# If not set, the certifi library will be used to load certificates.
CUSTOM_CA_BUNDLE_PATH = "<your_custom_ca_bundle_file_path>"


def producer_connecting_to_custom_endpoint():
    producer_client = EventHubProducerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        eventhub_name=EVENTHUB_NAME,
        credential=DefaultAzureCredential(),
        custom_endpoint_address=CUSTOM_ENDPOINT_ADDRESS,
        connection_verify=CUSTOM_CA_BUNDLE_PATH,
    )

    with producer_client:
        # Without specifying partition_id or partition_key
        # the events will be distributed to available partitions via round-robin.
        event_data_batch = producer_client.create_batch()
        event_data_batch.add(EventData("Single message"))
        producer_client.send_batch(event_data_batch)
        print("Send a message.")


def on_event(partition_context, event):
    # Put your code here.
    # If the operation is i/o intensive, multi-thread will have better performance.
    print("Received event from partition: {}.".format(partition_context.partition_id))


def consumer_connecting_to_custom_endpoint():
    consumer_client = EventHubConsumerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        consumer_group="$Default",
        eventhub_name=EVENTHUB_NAME,
        credential=DefaultAzureCredential(),
        custom_endpoint_address=CUSTOM_ENDPOINT_ADDRESS,
        connection_verify=CUSTOM_CA_BUNDLE_PATH,
    )

    try:
        with consumer_client:
            consumer_client.receive(
                on_event=on_event,
                starting_position="-1",  # "-1" is from the beginning of the partition.
            )
    except KeyboardInterrupt:
        print("Stopped receiving.")


producer_connecting_to_custom_endpoint()
consumer_connecting_to_custom_endpoint()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

# This EventHubs resource preparer is for future use after the tests are all compatible with
# The resource preparer.

import functools

from azure.mgmt.eventhub import EventHubManagementClient
from azure.mgmt.eventhub.models import Eventhub, AccessRights

from devtools_testutils import ResourceGroupPreparer, AzureMgmtPreparer, AzureTestError, FakeResource

from devtools_testutils.resource_testcase import RESOURCE_GROUP_PARAM

EVENTHUB_DEFAULT_AUTH_RULE_NAME = "RootManageSharedAccessKey"
EVENTHUB_NAMESPACE_PARAM = "eventhub_namespace"
EVENTHUB_PARAM = "eventhub"
EVENTHUB_AUTHORIZATION_RULE_PARAM = "eventhub_authorization_rule"


# Service Bus Namespace Preparer and its shorthand decorator
class EventHubNamespacePreparer(AzureMgmtPreparer):
    def __init__(
        self,
        name_prefix="",
        use_cache=False,
        sku="Standard",
        location="westus",
        parameter_name=EVENTHUB_NAMESPACE_PARAM,
        resource_group_parameter_name=RESOURCE_GROUP_PARAM,
        disable_recording=True,
        playback_fake_resource=None,
        client_kwargs=None,
        random_name_enabled=True,
    ):
        super(EventHubNamespacePreparer, self).__init__(
            name_prefix,
            24,
            random_name_enabled=random_name_enabled,
            disable_recording=disable_recording,
            playback_fake_resource=playback_fake_resource,
            client_kwargs=client_kwargs,
        )
        self.location = location
        self.sku = sku
        self.resource_group_parameter_name = resource_group_parameter_name
        self.parameter_name = parameter_name
        self.connection_string = ""
        if random_name_enabled:
            self.resource_moniker = self.name_prefix + "ehname"

        self.set_cache(use_cache, sku, location)

    def create_resource(self, name, **kwargs):
        if self.is_live:
            self.client = self.create_mgmt_client(EventHubManagementClient)  # type: EventHubManagementClient
            group = self._get_resource_group(**kwargs)
            namespace_async_operation = self.client.namespaces.begin_create_or_update(
                group.name,
                name,
                {
                    "sku": {"name": self.sku},
                    "location": self.location,
                },
            )
            self.resource = namespace_async_operation.result()

            key = self.client.namespaces.list_keys(group.name, name, EVENTHUB_DEFAULT_AUTH_RULE_NAME)
            self.connection_string = key.primary_connection_string
            self.key_name = key.key_name
            self.primary_key = key.primary_key
        else:
            self.resource = FakeResource(name=name, id=name)
            self.connection_string = "Endpoint=sb://{}.servicebus.windows.net/;SharedAccessKeyName=test;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX=".format(
                name
            )
            self.key_name = EVENTHUB_DEFAULT_AUTH_RULE_NAME
            self.primary_key = "ZmFrZV9hY29jdW50X2tleQ=="
        return {
            self.parameter_name: self.resource,
            "{}_connection_string".format(self.parameter_name): self.connection_string,
            "{}_key_name".format(self.parameter_name): self.key_name,
            "{}_primary_key".format(self.parameter_name): self.primary_key,
            "{}_management_client".format(self.parameter_name): self.client,
        }

    def remove_resource(self, name, **kwargs):
        if self.is_live:
            group = self._get_resource_group(**kwargs)
            self.client.namespaces.begin_delete(group.name, name).wait()

    def _get_resource_group(self, **kwargs):
        try:
            return kwargs.get(self.resource_group_parameter_name)
        except KeyError:
            template = (
                "To create an event hub a resource group is required. Please add "
                "decorator @{} in front of this event hub preparer."
            )
            raise AzureTestError(template.format(ResourceGroupPreparer.__name__))


# Shared base class for event hub sub-resources that require a namespace and RG to exist.
class _EventHubChildResourcePreparer(AzureMgmtPreparer):
    def __init__(
        self,
        name_prefix="",
        resource_group_parameter_name=RESOURCE_GROUP_PARAM,
        eventhub_namespace_parameter_name=EVENTHUB_NAMESPACE_PARAM,
        disable_recording=True,
        playback_fake_resource=None,
        client_kwargs=None,
        random_name_enabled=True,
    ):
        super(_EventHubChildResourcePreparer, self).__init__(
            name_prefix,
            24,
            random_name_enabled=random_name_enabled,
            disable_recording=disable_recording,
            playback_fake_resource=playback_fake_resource,
            client_kwargs=client_kwargs,
        )
        self.resource_group_parameter_name = resource_group_parameter_name
        self.eventhub_namespace_parameter_name = eventhub_namespace_parameter_name

    def _get_resource_group(self, **kwargs):
        try:
            return kwargs.get(self.resource_group_parameter_name)
        except KeyError:
            template = (
                "To create this event hub child resource event hub a resource group is required. Please add "
                "decorator @{} in front of this event hub preparer."
            )
            raise AzureTestError(template.format(ResourceGroupPreparer.__name__))

    def _get_namespace(self, **kwargs):
        try:
            return kwargs.get(self.eventhub_namespace_parameter_name)
        except KeyError:
            template = (
                "To create this event hub child resource a event hub namespace is required. Please add "
                "decorator @{} in front of this event hub preparer."
            )
            raise AzureTestError(template.format(EventHubNamespacePreparer.__name__))


class EventHubPreparer(_EventHubChildResourcePreparer):
    def __init__(
        self,
        name_prefix="",
        use_cache=False,
        parameter_name=EVENTHUB_PARAM,
        resource_group_parameter_name=RESOURCE_GROUP_PARAM,
        eventhub_namespace_parameter_name=EVENTHUB_NAMESPACE_PARAM,
        disable_recording=True,
        playback_fake_resource=None,
        client_kwargs=None,
        random_name_enabled=True,
    ):
        super(EventHubPreparer, self).__init__(
            name_prefix,
            random_name_enabled=random_name_enabled,
            resource_group_parameter_name=resource_group_parameter_name,
            eventhub_namespace_parameter_name=eventhub_namespace_parameter_name,
            disable_recording=disable_recording,
            playback_fake_resource=playback_fake_resource,
            client_kwargs=client_kwargs,
        )
        self.parameter_name = parameter_name
        if random_name_enabled:
            self.resource_moniker = self.name_prefix + "eventhub"
        self.set_cache(use_cache)

    def create_resource(self, name, **kwargs):
        if self.is_live:
            self.client = self.create_mgmt_client(EventHubManagementClient)
            group = self._get_resource_group(**kwargs)
            namespace = self._get_namespace(**kwargs)
            self.resource = self.client.event_hubs.create_or_update(group.name, namespace.name, name, {})
        else:
            self.resource = FakeResource(name=name, id=name)
        return {
            self.parameter_name: self.resource,
        }

    def remove_resource(self, name, **kwargs):
        if self.is_live:
            group = self._get_resource_group(**kwargs)
            namespace = self._get_namespace(**kwargs)
            self.client.event_hubs.delete(group.name, namespace.name, name, polling=False)


class EventHubNamespaceAuthorizationRulePreparer(_EventHubChildResourcePreparer):
    def __init__(
        self,
        name_prefix="",
        use_cache=False,
        access_rights=[AccessRights.manage, AccessRights.send, AccessRights.listen],
        parameter_name=EVENTHUB_AUTHORIZATION_RULE_PARAM,
        resource_group_parameter_name=RESOURCE_GROUP_PARAM,
        eventhub_namespace_parameter_name=EVENTHUB_NAMESPACE_PARAM,
        disable_recording=True,
        playback_fake_resource=None,
        client_kwargs=None,
        random_name_enabled=True,
    ):
        super(EventHubNamespaceAuthorizationRulePreparer, self).__init__(
            name_prefix,
            random_name_enabled=random_name_enabled,
            resource_group_parameter_name=resource_group_parameter_name,
            eventhub_namespace_parameter_name=eventhub_namespace_parameter_name,
            disable_recording=disable_recording,
            playback_fake_resource=playback_fake_resource,
            client_kwargs=client_kwargs,
        )
        self.parameter_name = parameter_name
        self.access_rights = access_rights
        if random_name_enabled:
            self.resource_moniker = self.name_prefix + "sbnameauth"
        self.set_cache(use_cache, access_rights)

    def create_resource(self, name, **kwargs):
        if self.is_live:
            self.client = self.create_mgmt_client(EventHubManagementClient)
            group = self._get_resource_group(**kwargs)
            namespace = self._get_namespace(**kwargs)
            self.resource = self.client.namespaces.create_or_update_authorization_rule(
                group.name, namespace.name, name, self.access_rights
            )

            key = self.client.namespaces.list_keys(group.name, namespace.name, name)
            connection_string = key.primary_connection_string
        else:
            self.resource = FakeResource(name=name, id=name)
            connection_string = "https://microsoft.com"
        return {
            self.parameter_name: self.resource,
            "{}_connection_string".format(self.parameter_name): connection_string,
        }

    def remove_resource(self, name, **kwargs):
        if self.is_live:
            group = self._get_resource_group(**kwargs)
            namespace = self._get_namespace(**kwargs)
            self.client.namespaces.delete_authorization_rule(group.name, namespace.name, name, polling=False)


class EventHubAuthorizationRulePreparer(_EventHubChildResourcePreparer):
    def __init__(
        self,
        name_prefix="",
        use_cache=False,
        access_rights=[AccessRights.manage, AccessRights.send, AccessRights.listen],
        parameter_name=EVENTHUB_AUTHORIZATION_RULE_PARAM,
        resource_group_parameter_name=RESOURCE_GROUP_PARAM,
        eventhub_namespace_parameter_name=EVENTHUB_NAMESPACE_PARAM,
        eventhub_parameter_name=EVENTHUB_PARAM,
        disable_recording=True,
        playback_fake_resource=None,
        client_kwargs=None,
        random_name_enabled=True,
    ):
        super(EventHubAuthorizationRulePreparer, self).__init__(
            name_prefix,
            random_name_enabled=random_name_enabled,
            resource_group_parameter_name=resource_group_parameter_name,
            eventhub_namespace_parameter_name=eventhub_namespace_parameter_name,
            disable_recording=disable_recording,
            playback_fake_resource=playback_fake_resource,
            client_kwargs=client_kwargs,
        )
        self.parameter_name = parameter_name
        self.access_rights = access_rights
        self.eventhub_parameter_name = eventhub_parameter_name
        if random_name_enabled:
            self.resource_moniker = self.name_prefix + "evnethubauth"
        self.set_cache(use_cache, access_rights)

    def create_resource(self, name, **kwargs):
        if self.is_live:
            self.client = self.create_mgmt_client(EventHubManagementClient)
            group = self._get_resource_group(**kwargs)
            namespace = self._get_namespace(**kwargs)
            eventhub = self._get_eventhub(**kwargs)
            self.resource = self.client.event_hubs.create_or_update_authorization_rule(
                group.name, namespace.name, eventhub.name, name, self.access_rights
            )

            key = self.client.event_hubs.list_keys(group.name, namespace.name, eventhub.name, name)
            connection_string = key.primary_connection_string
        else:
            self.resource = FakeResource(name=name, id=name)
            connection_string = "https://microsoft.com"
        return {
            self.parameter_name: self.resource,
            "{}_connection_string".format(self.parameter_name): connection_string,
        }

    def remove_resource(self, name, **kwargs):
        if self.is_live:
            group = self._get_resource_group(**kwargs)
            namespace = self._get_namespace(**kwargs)
            eventhub = self._get_eventhub(**kwargs)
            self.client.event_hubs.delete_authorization_rule(
                group.name, namespace.name, eventhub.name, name, polling=False
            )

    def _get_eventhub(self, **kwargs):
        try:
            return kwargs.get(self.eventhub_parameter_name)
        except KeyError:
            template = (
                "To create this event hub authorization rule a event hub is required. Please add "
                "decorator @{} in front of this event hub preparer."
            )
            raise AzureTestError(template.format(EventHubPreparer.__name__))


CachedEventHubNamespacePreparer = functools.partial(EventHubNamespacePreparer, use_cache=True)
CachedEventHubPreparer = functools.partial(EventHubPreparer, use_cache=True)
